# sacher_epos.py, python wrapper for sacher epos motor
# David Christle <christle@uchicago.edu>, August 2014
#

"""
This is the actual version that works
But only in the lab32 virtual environment
"""

#from instrument import Instrument
import visa
import types
import numpy
#import qt

import ctypes
import numpy as np
import logging
#from instrument import Instrument
from ctypes.wintypes import DWORD
from ctypes.wintypes import WORD
import ctypes.wintypes
import time

"""
okay so we import a bunch of random stuff
I always forget what ctypes is for but I'll worry about it later
"""

#from subprocess import Popen, PIPE
#from multiprocessing.managers import BaseManager
#import atexit
#import os

#python32_dir = "C:\\Users\\Alex\\Miniconda3\\envs\\lab32"

#assert os.path.isdir(python32_dir)
#os.chdir(python32_dir)

#derp = "C:\\Users\\Alex\\Documents\\wow_such_code"

#assert os.path.isdir(derp)
#os.chdir(derp)

#p = Popen([python32_dir + "\\python.exe", derp + "\\delegate.py"], stdout=PIPE, cwd=derp)

#atexit.register(p.terminate)

#port = int(p.stdout.readline())
#authkey = p.stdout.read()

#print(port, authkey)

#m = BaseManager(address=("localhost", port), authkey=authkey)
#m.connect()

# tell manager to expect an attribute called LibC
#m.register("SacherLasaTeknique")

# access and use libc
#libc = m.SacherLasaTeknique()
#print(libc.vcs())

#eposlib = ctypes.windll.eposcmd



eposlib = ctypes.windll.LoadLibrary('C:\\Users\\Carbro\\Desktop\\Charmander\\EposCmd.dll')

DeviceName = b'EPOS'
ProtocolStackName = b'MAXON_RS232'
InterfaceName = b'RS232'

"""
Max on
Max off
but anyway it looks like ctypes is the thing that's talking to the epos dll
"""


HISTCHAN = 65536
TTREADMAX = 131072
RANGES = 8

MODE_HIST = 0
MODE_T2 = 2
MODE_T3 = 3

FLAG_OVERFLOW = 0x0040
FLAG_FIFOFULL = 0x0003

# in mV
ZCMIN = 0
ZCMAX = 20
DISCRMIN = 0
DISCRMAX = 800

# in ps
OFFSETMIN = 0
OFFSETMAX = 1000000000

# in ms
ACQTMIN = 1
ACQTMAX = 10*60*60*1000

# in mV
PHR800LVMIN = -1600
PHR800LVMAX = 2400

"""
wooooooo a bunch a variables and none of them are explained
way to go dc you da real champ
"""


class Sacher_EPOS():

    """
    ok before I dive into this giant Sacher class thing let me just list here all the functions that are being defined in this class:
    check(self)
        before
        wreck(self)

    ok but actually:
    __init__(self, name, address, reset=False)
    __del__(self)
    get_bit(self, byteval,idx)
    _u32todouble(self, uinput)
    open(self)
    close(self)
    get_offset(self)
    fine_tuning_steps(self, steps)
    set_new_offset(self, new_offset)
    get_motor_position(self)
    set_target_position(self, target, absolute, immediately)
    do_get_wavelength(self)
    do_set_wavelength(self, wavelength)
    is_open(self)
    clear_fault(self)
    initialize(self)

    The last one is really long
    And also damn there are 16 of them
    I'll comment about them as I go through them
    """

    def __init__(self, name, address, reset=False):
        # Instrument.__init__(self, name, tags=['physical'])
        # self._port_name = str(address)
        self._port_name = address
        self._is_open = False
        self._HPM = True

        # self.add_parameter('wavelength',
        #     flags = Instrument.FLAG_GETSET,
        #     type = types.FloatType,
        #     units = 'nm',
        #     minval=1070.0,maxval=1180.0)

        # self.add_function('open')
        # self.add_function('close')
        # self.add_function('fine_tuning_steps')
        # self.add_function('get_motor_position')
        # self.add_function('set_target_position')
        #try:
        self.open()
        self.initialize()
        #except:
        #    logging.error('Error loading Sacher EPOS motor. In use?')
    """
    I mean to me this really seems like the initialize function
    so I wonder what initialize(self) is doing
    At any rate there doesn't seem to be a lot going on here
    """


    def __del__(self):
        # execute disconnect
        self.close()
        return
    """
    this might be the only self explanatory one
    it disconnects
    """

    @staticmethod
    def get_bit(byteval, idx):
    # def get_bit(self, byteval,idx):
        return ((byteval&(1<< idx ))!=0)
    """
    you get the bits, and then you use them
    but honestly I don't really get what this is doing
    sudo git a_clue
    """

    @staticmethod
    def _u32todouble(uinput):
    # def _u32todouble(self, uinput):
        # this function implements the really weird/non-standard U32 to
        # floating point conversion in the sacher VIs

        # get sign of number
        sign = Sacher_EPOS.get_bit(uinput,31)
        if sign == False:
            mantissa_sign = 1
        elif sign == True:
            mantissa_sign = -1
        exp_mask =  0b111111
        #print 'uin u is %d' % uinput
        #print 'type uin %s' % type(uinput)
        #print 'binary input is %s' % bin(long(uinput))
        # get sign of exponent
        if Sacher_EPOS.get_bit(uinput,7) == False:
            exp_sign = 1
        elif Sacher_EPOS.get_bit(uinput,7) == True:
            exp_sign = -1


        #print 'exp extract %s' % bin(int(uinput & exp_mask))
        #print 'exp conv %s' % (exp_sign*int(uinput & exp_mask))
        #print 'sign of exponent %s' % self.get_bit(uinput,7)
        #print 'binary constant is %s' % bin(int(0b10000000000000000000000000000000))
        mantissa_mask = 0b01111111111111111111111100000000
        # mantissa_mask = 0b0111111111111111111111110000000


        #print 'mantissa extract is %s' % bin((uinput & mantissa_mask) >> 8)
        mantissa = 1.0/1000000.0*float(mantissa_sign)*float((uinput & mantissa_mask) >> 8)
        #print 'mantissa is %.12f' % mantissa
        # print(1 if Sacher_EPOS.get_bit(uinput,31) else 0, mantissa, 1 if Sacher_EPOS.get_bit(uinput,7) else 0, uinput & exp_mask)
        output = mantissa*2.0**(float(exp_sign)*float(int(uinput & exp_mask)))
        #print 'output is %s' % output
        return output
    """
    ok dc gave some slight explanations here
    Apparently there's a "really weird/non-standard U32 to floating point conversion in the sacher VIs"
    It'd be gr8 if I knew what U32's were
    unsigned 32 bit something something?
    ah whatever
    I'll have to worry about this later
    """

    @staticmethod
    def _doubletou32(dinput):
        mantissa_bit = 0 if int(dinput / abs(dinput)) > 0 else 1
        exp_bit = 1 if -1 < dinput < 1 else 0

        b = np.ceil(np.log10(abs(dinput)))
        a = dinput / 10 ** b
        if dinput < 0:
            a = -a
        # print('a:\t{}\tb:\t{}'.format(a, b))

        d = np.log2(10) * b
        d_ = np.ceil(d)
        c = a * 2 ** (d - d_)
        # print('c:\t{}\td_:{}\toriginal:\t{}'.format(c, d_, c * 2 ** d_))

        return (int(mantissa_bit) << 31) + (int(c * 1e6) << 8) + (int(exp_bit) << 7) + int(abs(d_))



    def open(self):
        eposlib.VCS_OpenDevice.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.POINTER(DWORD)]
        eposlib.VCS_OpenDevice.restype = ctypes.wintypes.HANDLE
        buf = ctypes.pointer(DWORD(0))
        ret = ctypes.wintypes.HANDLE()

        #print 'types are all %s %s %s %s %s' % (type(DeviceName), type(ProtocolStackName), type(InterfaceName), type(self._port_name), type(buf))
        ret = eposlib.VCS_OpenDevice(DeviceName, ProtocolStackName, InterfaceName, self._port_name, buf)
        self._keyhandle = ret
        #print 'keyhandle is %s' % self._keyhandle
        #print 'open device ret %s' % buf
        #print 'printing'
        #print buf.contents.value
        #print 'done printer'
        if int(buf.contents.value) >= 0:
            self._is_open = True
            self._keyhandle = ret
        return
    """
    I have absolutely no idea what the hell this is doing
    Considering that close(self) is apparently closing the EPOS motor, maybe this is opening it
    """


    def close(self):
        print('closing EPOS motor.')

        eposlib.VCS_CloseDevice.argtypes = [ctypes.wintypes.HANDLE, ctypes.POINTER(DWORD)]
        eposlib.VCS_CloseDevice.restype = ctypes.wintypes.BOOL
        buf = ctypes.pointer(DWORD(0))
        ret = ctypes.wintypes.BOOL()

        ret = eposlib.VCS_CloseDevice(self._keyhandle, buf)

        #print 'close device returned %s' % buf

        if int(buf.contents.value) >= 0:
            self._is_open = False
        else:
            logging.error(__name__ + ' did not close Sacher EPOS motor correctly.')
        return
    """
    Apparently this closes the EPOS motor
    I don't know what "opening" and "closing" the motor means though
    and yeah also these random variables don't make any sense to me
    """



    def get_motor_current(self):
        nodeID = ctypes.wintypes.WORD(0)
        eposlib.VCS_GetCurrentIs.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetCurrentIs.restype = ctypes.wintypes.BOOL

        motorCurrent = ctypes.c_uint8(0)
        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_GetCurrentIs(self._keyhandle, nodeID, ctypes.byref(motorCurrent), ctypes.byref(buf))
        return motorCurrent.value

    """
    Not sure what this is doing yet
    """



    def find_home(self):
        nodeID = ctypes.wintypes.WORD(0)
        eposlib.VCS_FindHome.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_FindHome.restype = ctypes.wintypes.BOOL

        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_FindHome(self._keyhandle, nodeID, ctypes.c_uint8(35), ctypes.byref(buf))
        print('Homing: {}'.format(ret))
        return ret

    """
    Not sure what this is doing yet
    """



    def restore(self):
        nodeID = ctypes.wintypes.WORD(0)
        eposlib.VCS_FindHome.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_FindHome.restype = ctypes.wintypes.BOOL

        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_Restore(self._keyhandle, nodeID, ctypes.byref(buf))
        print('Restore: {}'.format(ret))
        return ret

    """
    Not sure what this is doing yet
    """



    def get_offset(self):
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL
        # These are hardcoded values I got from the LabVIEW program -- I don't think
        # any documentation exists on particular object indices
        StoredPositionObject = ctypes.wintypes.WORD(8321)
        StoredPositionObjectSubindex = ctypes.c_uint8(0)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_int32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))

        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_int32))
        if ret == 0:
            logging.error(__name__ + ' Could not read stored position from Sacher EPOS motor')
        return CastedObjectData[0]

    """
    Not sure what this is doing yet
    """



    def fine_tuning_steps(self, steps):
        current_motor_pos = self.get_motor_position()
        self._offset = self.get_offset()
        self.set_target_position(steps, False, True)
        new_motor_pos = self.get_motor_position()
        #print('New motor position is %s' % new_motor_pos)
        #print 'new offset is %s' % (new_motor_pos-current_motor_pos+self._offset)
        self.set_new_offset(new_motor_pos-current_motor_pos+self._offset)

    """
    Not sure what this is doing yet
    """



    def set_new_offset(self, new_offset):
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        eposlib.VCS_SetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_SetObject.restype = ctypes.wintypes.BOOL
        #print 'setting new offset'

        StoredPositionObject = ctypes.wintypes.WORD(8321)
        StoredPositionObjectSubindex = ctypes.c_uint8(0)
        StoredPositionNbBytesToWrite = ctypes.wintypes.DWORD(4)

        ObjectDataArray = (ctypes.c_uint32*1)(new_offset)
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesWritten = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_SetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToWrite, StoredPositionNbBytesWritten, ctypes.byref(buf))

        if ret == 0:
            logging.error(__name__ + ' Could not write stored position from Sacher EPOS motor')
        return

    """
    Not sure what this is doing yet
    """



    def set_coeffs(self, a, b, c, min_wl, max_wl):
        print('')
        print("setting coefficients...")
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        eposlib.VCS_SetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_SetObject.restype = ctypes.wintypes.BOOL
        #print 'setting new offset'
        d = (min_wl << 16) + max_wl

        StoredPositionObject = ctypes.wintypes.WORD(8204)
        for subidx, coeff in enumerate([a, b, c]):
            print(subidx, coeff)
            StoredPositionObjectSubindex = ctypes.c_uint8(subidx + 1)
            StoredPositionNbBytesToWrite = ctypes.wintypes.DWORD(4)


            ObjectDataArray = (ctypes.c_uint32*1)(self._doubletou32(coeff))
            ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
            StoredPositionNbBytesWritten = ctypes.pointer(ctypes.wintypes.DWORD(0))
            ret = eposlib.VCS_SetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToWrite, StoredPositionNbBytesWritten, ctypes.byref(buf))

        StoredPositionObjectSubindex = ctypes.c_uint8(4)
        StoredPositionNbBytesToWrite = ctypes.wintypes.DWORD(4)
        ObjectDataArray = (ctypes.c_uint32*1)(d)
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesWritten = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_SetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToWrite, StoredPositionNbBytesWritten, ctypes.byref(buf))

        print('Coefficients are %s %s %s' % (self._doubleA, self._doubleB, self._doubleC))

        if ret == 0:
            logging.error(__name__ + ' Could not write stored position from Sacher EPOS motor')
        return

    """
    Not sure what this is doing yet
    """



    def get_motor_position(self):
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        pPosition = ctypes.pointer(ctypes.c_long())
        eposlib.VCS_GetPositionIs.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.c_long), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetPositionIs.restype = ctypes.wintypes.BOOL
        ret = eposlib.VCS_GetPositionIs(self._keyhandle, nodeID, pPosition, ctypes.byref(buf))
        #print 'get motor position ret %s' % ret
        #print 'get motor position buf %s' % buf.value
        #print 'get motor position value %s' % pPosition.contents.value

        return pPosition.contents.value

        #print('getting motor position...')
        #print(ret)
        #return print(pPosition.contents.value)

    """
    Not sure what this is doing yet
    """



    def set_target_position(self, target, absolute, immediately):
        #print('check #1')

        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        # First, set enabled state

        # print('#5 Motor current: {}'.format(self.get_motor_current()))
        # print('#5 Motor current: {}'.format(self.get_motor_current()))
        # print('#5 Motor current: {}'.format(self.get_motor_current()))
        # print('#5 Motor current: {}'.format(self.get_motor_current()))
        # print('#5 Motor current: {}'.format(self.get_motor_current()))

        ret = eposlib.VCS_SetEnableState(self._keyhandle,nodeID,ctypes.byref(buf))
        #print('Enable state ret %s buf %s' % (ret, buf.value))

        # print('#6 Motor current: {}'.format(self.get_motor_current()))
        # print('#6 Motor current: {}'.format(self.get_motor_current()))
        # print('#6 Motor current: {}'.format(self.get_motor_current()))
        # print('#6 Motor current: {}'.format(self.get_motor_current()))
        # print('#6 Motor current: {}'.format(self.get_motor_current()))

        pTarget = ctypes.c_long(target)
        pAbsolute = ctypes.wintypes.BOOL(absolute)
        pImmediately = ctypes.wintypes.BOOL(immediately)

        eposlib.VCS_MoveToPosition.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.c_long, ctypes.wintypes.BOOL, ctypes.wintypes.BOOL, ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_MoveToPosition.restype = ctypes.wintypes.BOOL

        #print('check #2')

        #print('About to set motor position')
        #print('Current motor position is %d' % (self.get_motor_position()))


        ret = eposlib.VCS_MoveToPosition(self._keyhandle, nodeID, pTarget, pAbsolute, pImmediately, ctypes.byref(buf))
        # print('#7 Motor current: {}'.format(self.get_motor_current()))
        # print('#7 Motor current: {}'.format(self.get_motor_current()))
        # print('#7 Motor current: {}'.format(self.get_motor_current()))
        # print('#7 Motor current: {}'.format(self.get_motor_current()))
        # print('#7 Motor current: {}'.format(self.get_motor_current()))

        #print('set motor position ret %s' % ret)
        #print('set motor position buf %s' % buf.value)

        steps_per_second = 14494.0 # hardcoded, estimated roughly, unused now

        nchecks = 0
        #print('check #3')
        while nchecks < 1000:
            # get the movement state. a movement state of 1 indicates the motor
            # is done moving
            # print('')
            # print('check #4')

            #print('Motor current: {}'.format(self.get_motor_current()))
            print('Motor position: {}'.format(self.get_motor_position()))
            #print('Motor offset: {}'.format(self.get_offset()))

            self._offset = self.get_offset()
            #print('Motor offset is %s' % self._offset)

            pMovementState = ctypes.pointer(ctypes.wintypes.BOOL())
            # print(pMovementState.contents.value)

            eposlib.VCS_GetMovementState.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.wintypes.BOOL), ctypes.POINTER(ctypes.wintypes.DWORD)]
            eposlib.VCS_GetMovementState.restype = ctypes.wintypes.BOOL
            # print('Getting movement state')
            ret = eposlib.VCS_GetMovementState(self._keyhandle, nodeID, pMovementState, ctypes.byref(buf))

            # print('set motor position ret %s' % ret)
            # print('set motor position buf %s' % buf.value)
            # print('Movement state is %s' % pMovementState.contents.value)
            if pMovementState.contents.value == 1:
                break
            nchecks = nchecks + 1
            # print('Current motor position is %d' % self.get_motor_position())
            # print('check #5')
            # print(nchecks)
            # print('')
            time.sleep(0.01)
        # Now set disabled state
        ret = eposlib.VCS_SetDisableState(self._keyhandle,nodeID,ctypes.byref(buf))
        #print('check #6')
        #print('Disable state ret %s buf %s' % (ret, buf.value))
        #print('Final motor position is %d' % (self.get_motor_position()))
        #print('check #7')
        return ret

    """
    Not sure what this is doing yet
    """



    def fuck_my_life(self, wavelength):
        print('goddamn this piece of shit')
        print('')

        print('Coefficients are %s %s %s' % (self._doubleA, self._doubleB, self._doubleC))
        #print('#3 Motor current: {}'.format(self.get_motor_current()))
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)

        # Step 1: Get the actual motor position

        #print('Getting motor position')

        current_motor_pos = self.get_motor_position()

        # Step 2: Get the motor offset

        self._offset = self.get_offset()
        #print('Motor offset is %s' % self._offset)

        # Step 3: Convert the desired wavelength into a position
        # Check sign of position-to-wavelength
        pos0 = self._doubleA*(0.0)**2.0 + self._doubleB*0.0 + self._doubleC
        pos5000 = self._doubleA*(5000.0)**2.0 + self._doubleB*5000.0 + self._doubleC
        #    logging.error(__name__ + ' Sacher wavelength calibration polynomials indicated a wrong wavelength direction')
        # If that's OK, use the quadratic formula to calculate the roots
        b2a = -1.0*self._doubleB/(2.0*self._doubleA)
        sqrtarg = self._doubleB**2.0/(4.0*self._doubleA**2.0) - (self._doubleC - wavelength)/self._doubleA
        # print('wut da fuuuu')
        # print(b2a)
        # print(sqrtarg)
        # print(pos0)
        # print(pos5000)
        if sqrtarg < 0.0:
            logging.error(__name__ + ' Negative value under square root sign -- something is wrong')
        if pos0 > pos5000:
            # Take the + square root solution
            x = b2a - np.sqrt(sqrtarg)
        elif pos0 < pos5000:
            x = b2a + np.sqrt(sqrtarg)

        print(b2a)
        print(np.sqrt(sqrtarg))
        #print('Position is %s' % x)

        wavelength_to_pos = int(round(x))
        # Step 4: Calculate difference between the output position and the stored offset

        #print('Step 4...')

        diff_wavelength_offset = wavelength_to_pos - int(self._offset)
        print('wavelength_to_pos: {}'.format(wavelength_to_pos))
        print('diff_wavelength_offset: {}'.format(diff_wavelength_offset))
        print('self._offset: {}'.format(int(self._offset)))

    """
    Not sure what this is doing yet
    """



    def do_get_wavelength(self):
        self._offset = self.get_offset()
        #self._currentwl = self._doubleA*(self._offset)**2.0 + self._doubleB*self._offset + self._doubleC
        self._currentwl = self._doubleA*(self.get_motor_position())**2.0 + self._doubleB * self.get_motor_position() + self._doubleC
        print('Current wavelength: %.3f nm' % self._currentwl)
        return self._currentwl

    """
    Not sure what this is doing yet
    """



    def do_set_wavelength(self, wavelength):
        print('setting wavelength...')
        print('')

        #print('Coefficients are %s %s %s' % (self._doubleA, self._doubleB, self._doubleC))
        #print('#3 Motor current: {}'.format(self.get_motor_current()))
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)

        # Step 1: Get the actual motor position

        #print('Getting motor position')

        current_motor_pos = self.get_motor_position()

        # Step 2: Get the motor offset

        self._offset = self.get_offset()
        #print('Motor offset is %s' % self._offset)

        # Step 3: Convert the desired wavelength into a position
        # Check sign of position-to-wavelength
        pos0 = self._doubleA*(0.0)**2.0 + self._doubleB*0.0 + self._doubleC
        pos5000 = self._doubleA*(5000.0)**2.0 + self._doubleB*5000.0 + self._doubleC
        #    logging.error(__name__ + ' Sacher wavelength calibration polynomials indicated a wrong wavelength direction')
        # If that's OK, use the quadratic formula to calculate the roots
        b2a = -1.0*self._doubleB/(2.0*self._doubleA)
        sqrtarg = self._doubleB**2.0/(4.0*self._doubleA**2.0) - (self._doubleC - wavelength)/self._doubleA
        # print('wut da fuuuu')
        # print(b2a)
        # print(sqrtarg)
        # print(pos0)
        # print(pos5000)
        if sqrtarg < 0.0:
            logging.error(__name__ + ' Negative value under square root sign -- something is wrong')
        if pos0 > pos5000:
            # Take the + square root solution
            x = b2a - np.sqrt(sqrtarg)
        elif pos0 < pos5000:
            x = b2a + np.sqrt(sqrtarg)
            #x is what the motor position should be

        #print('Position is %s' % x)

        wavelength_to_pos = int(round(x))
        # Step 4: Calculate difference between the output position and the stored offset

        #print('Step 4...')

        diff_wavelength_offset = wavelength_to_pos - int(self._offset)

        #print('Diff wavelength offset %s' % diff_wavelength_offset)

        # Step 5: If HPM is activated and the wavelength position is lower, overshoot
        # the movement by 10,000 steps

        #print('Step 5...')
        #print('#4 Motor current: {}'.format(self.get_motor_current()))

        if 1==2:
            print('uh-oh')
        # if self._HPM and diff_wavelength_offset < 0:
        #
        #     print('Overshooting by 10000')
        #
        #     self.set_target_position(diff_wavelength_offset - 10000, False, True)
        #     # Step 6: Set the real target position
        #
        #     """
        #     HEY LOOK EVERYONE RIGHT ABOVE HERE THIS IS THE STUPID THING THAT'S NOT WORKING!
        #     """
        #
        #     #print('Step 6a... diff wavelength')
        #
        #     self.set_target_position(10000, False, True)
        else:

            #print('Step 6b... diff wavelength')

            #self.set_target_position(diff_wavelength_offset, False, True)
            """WRONG"""

            self.set_target_position(wavelength_to_pos, True, True)
            """this is the real shit right here

            I need to set the absolute position to true
            """

            #self.set_target_position(10000, False, True)


        # Step 7: Get the actual motor position
        new_motor_pos = self.get_motor_position()

        #print('New motor position is %s' % new_motor_pos)
        #print('new offset is %s' % (new_motor_pos-current_motor_pos+self._offset))

        self.set_new_offset(new_motor_pos-current_motor_pos+self._offset)

        # Step 8, get and print current wavelength

        #print('Current wavelength is %.3f' % self.do_get_wavelength())

        #print('setting wavelength done')
        return

    """
    Not sure what this is doing yet
    """



    def is_open(self):
        return self._is_open

    """
    Not sure what this is doing yet
    """



    def clear_fault(self):
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_ClearFault(self._keyhandle,nodeID,ctypes.byref(buf))
        print('clear fault buf %s, ret %s' % (buf, ret))
        if ret == 0:
            errbuf = ctypes.create_string_buffer(64)
            eposlib.VCS_GetErrorInfo(buf, errbuf, WORD(64))
            raise ValueError(errbuf.value)

    """
    Not sure what this is doing yet
    """



    def initialize(self):
        nodeID = ctypes.wintypes.WORD(0)
        buf = ctypes.wintypes.DWORD(0)
        BaudRate = DWORD(38400)
        Timeout = DWORD(100)
        ret = eposlib.VCS_SetProtocolStackSettings(self._keyhandle,BaudRate,Timeout,ctypes.byref(buf))
        #print 'set protocol buf %s ret %s' % (buf, ret)
        if ret == 0:
            errbuf = ctypes.create_string_buffer(64)
            #eposlib.VCS_GetErrorInfo(buf, errbuf, WORD(64))
            raise ValueError(errbuf.value)

        buf = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_ClearFault(self._keyhandle,nodeID,ctypes.byref(buf))
        #print 'clear fault buf %s, ret %s' % (buf, ret)
        if ret == 0:
            errbuf = ctypes.create_string_buffer(64)
            eposlib.VCS_GetErrorInfo(buf, errbuf, WORD(64))
            raise ValueError(errbuf.value)
        buf = ctypes.wintypes.DWORD(0)
        plsenabled = ctypes.wintypes.DWORD(0)
        ret = eposlib.VCS_GetEnableState(self._keyhandle,nodeID,ctypes.byref(plsenabled),ctypes.byref(buf))
        #print 'get enable state buf %s ret %s and en %s' % (buf, ret, plsenabled)
        if ret == 0:
            errbuf = ctypes.create_string_buffer(64)
            eposlib.VCS_GetErrorInfo(buf, errbuf, WORD(64))
            raise ValueError(errbuf.value)

        if int(plsenabled.value) != 0:
            logging.warning(__name__ + ' EPOS motor enabled, disabling before proceeding.')
            ret = eposlib.VCS_SetDisableState(self._keyhandle,nodeID,ctypes.byref(buf))
            if int(ret) != 0:
                logging.warning(__name__ + ' EPOS motor successfully disabled, proceeding')
            else:
                logging.error(__name__ + ' EPOS motor was not successfully disabled!')
        buf = ctypes.wintypes.DWORD(0)
        Counts = WORD(512) # incremental encoder counts in pulses per turn
        PositionSensorType = WORD(4)
        ret = eposlib.VCS_SetEncoderParameter(self._keyhandle,nodeID,Counts,PositionSensorType,ctypes.byref(buf))

##        if ret == int(0):
##            print 'errr'
##            errbuf = ctypes.create_string_buffer(64)
##            print 'sending'
##            eposlib.VCS_GetErrorInfo.restype = ctypes.wintypes.BOOL
##            print 'boolerrorinfo'
##            eposlib.VCS_GetErrorInfo.argtypes = [ctypes.wintypes.DWORD, ctypes.c_char_p, ctypes.wintypes.WORD]
##            print 'arg'
##
##            ret = eposlib.VCS_GetErrorInfo(buf, ctypes.byref(errbuf), WORD(64))
##            print 'err'
##            raise ValueError(errbuf.value)
        # For some reason, it appears normal in the LabVIEW code that this
        # function actually returns an error, i.e. the return value is zero
        # and the buffer has a non-zero error code in it; the LabVIEW code
        # doesn't check it.
        # Also, it appears that in the 2005 version of this DLL, the function
        # VCS_GetErrorInfo doesn't exist!

        # Get operation mode, check if it's 1 -- this is "profile position mode"
        buf = ctypes.wintypes.DWORD(0)
        pMode = ctypes.pointer(ctypes.c_int8())
        eposlib.VCS_GetOperationMode.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.c_int8), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetOperationMode.restype = ctypes.wintypes.BOOL
        ret = eposlib.VCS_GetOperationMode(self._keyhandle, nodeID, pMode, ctypes.byref(buf))
        # if mode is not 1, make it 1
        if pMode.contents.value != 1:
            eposlib.VCS_SetOperationMode.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.c_int8, ctypes.POINTER(ctypes.wintypes.DWORD)]
            eposlib.VCS_SetOperationMode.restype = ctypes.wintypes.BOOL
            pMode_setting = ctypes.c_int8(1)
            ret = eposlib.VCS_SetOperationMode(self._keyhandle, nodeID, pMode_setting, ctypes.byref(buf))
        eposlib.VCS_GetPositionProfile.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetPositionProfile.restype = ctypes.wintypes.BOOL
        pProfileVelocity = ctypes.pointer(ctypes.wintypes.DWORD())
        pProfileAcceleration = ctypes.pointer(ctypes.wintypes.DWORD())
        pProfileDeceleration = ctypes.pointer(ctypes.wintypes.DWORD())

        ret = eposlib.VCS_GetPositionProfile(self._keyhandle, nodeID, pProfileVelocity, pProfileAcceleration, pProfileDeceleration,ctypes.byref(buf))

        print(pProfileVelocity.contents.value, pProfileAcceleration.contents.value, pProfileDeceleration.contents.value)

        if (int(pProfileVelocity.contents.value) > int(11400) or int(pProfileAcceleration.contents.value) > int(60000) or int(pProfileDeceleration.contents.value) > int(60000)):
            eposlib.VCS_GetPositionProfile.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.DWORD, ctypes.wintypes.DWORD, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD)]
            eposlib.VCS_GetPositionProfile.restype = ctypes.wintypes.BOOL
            pProfileVelocity = ctypes.wintypes.DWORD(429)
            pProfileAcceleration = ctypes.wintypes.DWORD(429)
            pProfileDeceleration = ctypes.wintypes.DWORD(429)
            logging.warning(__name__ + ' GetPositionProfile out of bounds, resetting...')
            ret = eposlib.VCS_SetPositionProfile(self._keyhandle, nodeID, pProfileVelocity, pProfileAcceleration, pProfileDeceleration,ctypes.byref(buf))

        # Now get the motor position (stored position offset)
        # from the device's "homposition" object

        self._offset = self.get_offset()

        # Now read the stored 'calculation parameters'
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL

        # More hardcoded values
        StoredPositionObject = ctypes.wintypes.WORD(8204)
        StoredPositionObjectSubindex = ctypes.c_uint8(1)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))
        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_uint32))

        self._coefA = CastedObjectData[0]
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL

        # Get coefficient B
        StoredPositionObject = ctypes.wintypes.WORD(8204)
        StoredPositionObjectSubindex = ctypes.c_uint8(2)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))
        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_uint32))

        self._coefB = CastedObjectData[0]
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL

        # These are hardcoded values I got from the LabVIEW program -- I don't think
        # any documentation exists on particular object indices
        StoredPositionObject = ctypes.wintypes.WORD(8204)
        StoredPositionObjectSubindex = ctypes.c_uint8(3)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))
        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_uint32))

        self._coefC = CastedObjectData[0]

        # Get coefficient D
        eposlib.VCS_GetObject.argtypes = [ctypes.wintypes.HANDLE, ctypes.wintypes.WORD, ctypes.wintypes.WORD, ctypes.c_uint8, ctypes.c_void_p, ctypes.wintypes.DWORD, ctypes.POINTER(ctypes.wintypes.DWORD), ctypes.POINTER(ctypes.wintypes.DWORD)]
        eposlib.VCS_GetObject.restype = ctypes.wintypes.BOOL

        # These are hardcoded values I got from the LabVIEW program -- I don't think
        # any documentation exists on particular object indices
        StoredPositionObject = ctypes.wintypes.WORD(8204)
        StoredPositionObjectSubindex = ctypes.c_uint8(4)
        StoredPositionNbBytesToRead = ctypes.wintypes.DWORD(4)
        ObjectData = ctypes.c_void_p()
        ObjectDataArray = (ctypes.c_uint32*1)()
        ObjectData = ctypes.cast(ObjectDataArray, ctypes.POINTER(ctypes.c_uint32))
        StoredPositionNbBytesRead = ctypes.pointer(ctypes.wintypes.DWORD(0))
        ret = eposlib.VCS_GetObject(self._keyhandle, nodeID, StoredPositionObject, StoredPositionObjectSubindex, ObjectData, StoredPositionNbBytesToRead, StoredPositionNbBytesRead, ctypes.byref(buf))
        # Cast the object data to uint32
        CastedObjectData = ctypes.cast(ObjectData, ctypes.POINTER(ctypes.c_uint32))

        self._coefD = CastedObjectData[0]
        #print 'coefficients are %s %s %s %s' % (self._coefA, self._coefB, self._coefC, self._coefD)
        self._doubleA = self._u32todouble(self._coefA)
        self._doubleB = self._u32todouble(self._coefB)
        self._doubleC = self._u32todouble(self._coefC)
        firstHalf = np.int16(self._coefD >> 16)
        secondHalf = np.int16(self._coefD & 0xffff)
        # Set the minimum and maximum wavelengths for the motor
        self._minwl = float(firstHalf)/10.0
        self._maxwl = float(secondHalf)/10.0
        # print 'first %s second %s' % (firstHalf, secondHalf)
        # This returns '10871' and '11859' for the Sacher, which are the correct
        # wavelength ranges in Angstroms
        #print 'Now calculate the current wavelength position:'
        self._currentwl = self._doubleA*(self._offset)**2.0 + self._doubleB*self._offset + self._doubleC
        print('Current wavelength: %.3f nm' % self._currentwl)
        print('initializing done')
        return True

    """
    Not sure what this is doing yet
    """

"""
Also we're done with the Sacher_EPOS() class at this point
"""




if __name__ == '__main__':
    epos = Sacher_EPOS(None, b'COM3')
    #epos.set_coeffs(8.34529e-12,8.49218e-5,1081.92,10840,11860)

    #epos.do_get_wavelength()

    #print('#1 Motor current: {}'.format(epos.get_motor_current()))
    # epos.do_get_wavelength()
    # print('motor position is...')
    # current_pos = epos.get_motor_position()
    # print('current position is {}'.format(current_pos))
    # new_pos = current_pos + 10000
    # epos.set_target_position(new_pos, True, True)
    # print(epos.get_motor_position())
    #print('#2 Motor current: {}'.format(epos.get_motor_current()))

    #epos.find_home()
    #epos.restore()
    #time.sleep(7)

    epos.do_set_wavelength(1151.5)

    #epos.do_get_wavelength()
    print('Motor current: {}'.format(epos.get_motor_current()))
    print('Motor position: {}'.format(epos.get_motor_position()))

"""
OTHER MISC. NOTES:

increasing wavelength:
causes the square to rotate left
causes base to move to the left when square is stuck in
causes screw to loosen
causes large gold base to tighten

decreasing wavelength:
there's an overshoot when lowering wavelength
causes the square to rotate right
causes base to move to the right when square is stuck in
causes screw to tighten
causes large gold base to loosen, and also unplug the motor

Also you don't need to explicitly run epos.initialize() because there's an __init__ function which contains epos.initialize()
"""

#womp the end



from lantz import Feat, DictFeat, Action
from lantz.messagebased import MessageBasedDriver

from collections import OrderedDict

from time import sleep


class SR7265(MessageBasedDriver):
    """Signal Recovery 7265
    DSP Lock-in Amplifier

    Author: P. Mintun
    Version: 0.0.1
    Date: 11/13/2015

    "Stay safe and happy locking in..."

    This driver assumes that the COMM settings of the 7265 are set up as
    follows:

    ADDRESS = N
    TERMINATOR = [EOI]
    TECH ECHO = DISABLED

    SDC=ADF 1 ENABLED
    """

    DEFAULTS = {
        'COMMON': {
            'write_termination': '\r\n',
            'read_termination': '\r\n'
        }
    }

    INT_EXT_REF = OrderedDict([
                   ('Internal', 0),
                   ('Rear External', 1),
                   ('External', 2)
                   ])

    TIME_CONSTANTS = OrderedDict([
                     ('10 us' , 0  ),
                     ('20 us' , 1  ),
                     ('40 us' , 2  ),
                     ('80 us' , 3  ),
                     ('160 us', 4  ),
                     ('320 us', 5  ),
                     ('640 us', 6  ),
                     ('5 ms'  , 7  ),
                     ('10 ms' , 8  ),
                     ('20 ms' , 9  ),
                     ('50 ms' , 10 ),
                     ('100 ms', 11 ),
                     ('200 ms', 12 ),
                     ('500 ms', 13 ),
                     ('1 s'   , 14 ),
                     ('2 s'   , 15 ),
                     ('5 s'   , 16 ),
                     ('10 s'  , 17 ),
                     ('20 s'  , 18 ),
                     ('50 s'  , 19 ),
                     ('100 s' , 20 ),
                     ('200 s' , 21 ),
                     ('500 s' , 22 ),
                     ('1 ks'  , 23 ),
                     ('2 ks'  , 24 ),
                     ('5 ks'  , 25 ),
                     ('10 ks' , 26 ),
                     ('20 ks' , 27 ),
                     ('50 ks' , 28 ),
                     ('100 ks', 29 ),
                     ])

    AC_GAINS = OrderedDict([
               ('0 dB', 0),
               ('10 dB', 1),
               ('20 dB', 2),
               ('30 dB', 3),
               ('40 dB', 4),
               ('50 dB', 5),
               ('60 dB', 6),
               ('70 dB', 7),
               ('80 dB', 8),
               ('90 dB', 9),
               ])

    SENSITIVITIES = OrderedDict([
                    ('2 nV',   1  ),
                    ('5 nV',   2  ),
                    ('10 nV',  3  ),
                    ('20 nV',  4  ),
                    ('50 nV',  5  ),
                    ('100 nV', 6  ),
                    ('200 nV', 7  ),
                    ('500 nV', 8  ),
                    ('1 uV',   9  ),
                    ('2 uV',   10 ),
                    ('5 uV',   11 ),
                    ('10 uV',  12 ),
                    ('20 uV',  13 ),
                    ('50 uV',  14 ),
                    ('100 uV', 15 ),
                    ('200 uV', 16 ),
                    ('500 uV', 17 ),
                    ('1 mV',   18 ),
                    ('2 mV',   19 ),
                    ('5 mV',   20 ),
                    ('10 mV',  21 ),
                    ('20 mV',  22 ),
                    ('50 mV',  23 ),
                    ('100 mV', 24 ),
                    ('200 mV', 25 ),
                    ('500 mV', 26 ),
                    ('1 V',    27 ),
                    ])

    def remove_null(self, value):
        value = value.replace('\x00', '')
        return value

    @Feat()
    def idn(self):
        """
        Returns current instrument identification.
        """
        return self.query('ID')

    @Feat(units='V')
    def x(self):
        """
        Read x value from lockin.
        """
        return float(self.remove_null(self.query('X.')))

    @Feat(units='V')
    def y(self):
        """
        Read y value from lockin.
        """
        return float(self.remove_null(self.query('Y.')))

    @Feat()
    def xy(self):
        """
        Read x and y values from lockin simultaneously.
        """
        return [float(value) for value in self.remove_null(self.query('XY.')).split(',')]

    @Feat(units='V')
    def magnitude(self):
        """
        Read signal magnitude from lockin.
        """
        return float(self.remove_null(self.query('MAG.')))

    @Feat(units='degree')
    def phase(self):
        """
        Read signal phase from lockin.
        """
        return float(self.remove_null(self.query('PHA.')))

    @Feat()
    def mag_phase(self):
        """
        Read signal magnitude and phase from lockin.
        """
        return [float(value) for value in self.remove_null(self.query('MP.')).split(',')]

    @Action()
    def autophase(self):
        """
        Adds an offset to the phase of the lockin to minimize the y-channel
        signal and maximize the x-channel signal.
        """
        return self.write('AQN')

    @Feat(values=TIME_CONSTANTS)
    def time_constant(self):
        """
        Returns current time constant setting (in seconds).
        """
        return float(self.query('TC'))

    @time_constant.setter
    def time_constant(self, time_const):
        self.write('TC{}'.format(time_const))
        return

    @Feat()
    def frequency(self):
        """
        Read current signal frequency.
        """
        return float(self.remove_null(self.query('FRQ.')))

    @Feat(limits=(0, 250e3))
    def oscillator_freq(self):
        """
        Read internal oscillator frequency.
        """
        return float(self.remove_null(self.query('OF.')))

    @oscillator_freq.setter
    def oscillator_freq(self, frequency):
        """
        Set internal oscillator frequency.
        """
        self.write('OF.{}'.format(frequency))

    @Feat(values=AC_GAINS)
    def gain(self):
        """
        Read current AC gain (dB).
        """
        return int(self.query('ACGAIN'))

    @gain.setter
    def gain(self, gain_value):
        """
        Set current AC gain (dB)
        """
        self.write('ACGAIN{}'.format(gain_value))



    @Feat(values=SENSITIVITIES)
    def sensitivity(self):
        """
        Gets sensitivity according to the SENSITIVITIES table.
        """
        return int(self.query('SEN'))

    @sensitivity.setter
    def sensitivity(self, sen_it):
        """
        Sets value of sensitivity as described by SENSITIVITIES.
        """
        self.write('SEN{}'.format(sen_it))

    @Feat()
    def voltage_mode(self):
        """
        Gets the voltage mode
        """
        return int(self.query('VMODE'))

    @voltage_mode.setter
    def voltage_mode(self, n):
        """
        Sets value of voltage mode
        From page 138 of the 7265 manual, they say:
            n   Input configuration
            0   Both inputs grounded (test mode)
            1   A input only
            2   -B input only
            3   A-B differential mode
        """
        self.write('VMODE{}'.format(n))

    @Action()
    def autosensitivity(self):
        """
        Runs an autosensitivity operation.

        The instrument adjusts its full-scale sensitivity so that the magnitude
        output lies between 30-90 percent of full-scale
        """
        a = self.write('AS')
        sleep(7.5)  # wait for operation to complete
        return a

    @Action()
    def autogain(self):
        """
        Set current AC gain to automatic.
        """
        return self.write('AUTOMATIC1')

    @Feat(values=INT_EXT_REF)
    def int_ext_ref(self):
        """
        Check if lockin is internally or externally referenced
        """
        return int(self.query('IE'))

    @int_ext_ref.setter
    def int_ext_ref(self, value):
        """
        Set lockin to be internal or external reference
        """
        return self.write('IE {}'.format(value))

    def status_byte(self):
        return int(self.query('ST'))

    def get_bit(self, byte, bit):
        return True if int('{0:b}'.format(byte).zfill(8)[-(bit + 1)]) else False

    @Feat(values={True: True, False: False})
    def reference_unlock(self):
        return self.get_bit(self.status_byte(), 3)

    @Feat(values={True: True, False: False})
    def overload(self):
        return self.get_bit(self.status_byte(), 4)

    @Action()
    def autophase(self):
        """
        Automatically detect offset
        """
        return self.write('AXO')

    @Feat(values={True: 1, False: 0})
    def x_offset_enabled(self):
        return int(self.query('XOF?').split(',')[0])

    @x_offset_enabled.setter
    def x_offset_enabled(self, val):
        return self.write('XOF {}'.format(val))

    @Feat(values={True: 1, False: 0})
    def y_offset_enabled(self):
        return int(self.query('YOF?').split(',')[0])

    @x_offset_enabled.setter
    def y_offset_enabled(self, val):
        return self.write('YOF {}'.format(val))

    @Feat()
    def x_offset(self):
        return int(self.query('XOF?').split(',')[1])

    @x_offset.setter
    def x_offset(self, val):
        return self.write('XOF 1 {}'.format(val))

    @Feat()
    def y_offset(self):
        return int(self.query('YOF?').split(',')[1])

    @y_offset.setter
    def y_offset(self, val):
        return self.write('YOF 1 {}'.format(val))

if __name__ == '__main__':
    with SR7265.via_gpib(7) as inst:
        print('The instrument identification is ' + inst.idn)

        print('Testing signal readings')
        print('Signal X: {}V'.format(inst.x))
        print('Signal Y: {}V'.format(inst.y))
        print('Signal magnitude: {}V'.format(inst.magnitude))
        print('Signal phase: {}degrees'.format(inst.phase))

        print('Testing full quadrature readings')
        print('X,Y: {}V'.format(list(inst.xy)))
        print('Magnitude, Phase: {}'.format(list(inst.mag_phase)))
        inst.autophase
        sleep(2)
        print('Magnitude, Phase: {}'.format(list(inst.mag_phase)))

        print('Testing frequency code')
        print('Ref f: {}Hz'.format(inst.frequency))
        inst.oscillator_freq = 137.0
        print('Ref f: {}Hz'.format(inst.frequency))
        inst.oscillator_freq = 17
        print('Ref f: {}Hz'.format(inst.frequency))
        #
        print('Internal External Reference check')
        print('Internal/ext reference: {}'.format(inst.int_ext_ref))
        inst.int_ext_ref = 'ext'
        print('Internal/ext reference: {}'.format(inst.int_ext_ref))
        inst.int_ext_ref = 'int'
        print('Internal/ext reference: {}'.format(inst.int_ext_ref))

        print('Time constant check')
        print('Int TC: {}'.format(inst.time_constant_integer))
        print('TC (sec): {}s'.format(inst.time_constant))
        inst.time_constant_integer = 15
        print('Int TC: {}'.format(inst.time_constant_integer))
        print('TC (sec): {}s'.format(inst.time_constant))
        inst.time_constant_integer = 10
        print('Int TC: {}'.format(inst.time_constant_integer))
        print('TC (sec): {}s'.format(inst.time_constant))

        print('AC Gain Check')
        print('AC Gain: {}'.format(inst.gain))
        inst.gain = '30dB'
        print('AC Gain: {}'.format(inst.gain))
        inst.autogain
        print('AC Gain: {}'.format(inst.gain))

        print('Sensitivity Check')
        print('Sen: {}'.format(inst.sensitivity))
        inst.sensitivity = '2e-8V'
        print('Sen: {}'.format(inst.sensitivity))
        inst.autosensitivity()
        print('Sen: {}'.format(inst.sensitivity))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.aa.aotf
    ~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers for an AOTF Controller


    Implementation Notes
    --------------------

    There are currently two (disconnected) ways of setting the power for each
    line: powerdb and power.

    Sources::

        - MDSnC Manual

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

#TODO: Implement calibrated power.

from lantz.core import MessageBasedDriver
from lantz.core.mfeats import BoolFeat, BoolDictFeat, QuantityDictFeat


class MDSnC(MessageBasedDriver):
    """MDSnC synthesizer for AOTF.nC
    """

    CHANNELS = list(range(8))

    DRIVER_TRUE = 1
    DRIVER_FALSE = 0

    #: Enable the main ouput.
    main_enabled = BoolFeat(None, 'I{}')

    #: Enable a single channel.
    enabled = BoolDictFeat(None, 'L{key}O{value}', keys=CHANNELS)

    #: RF frequency for a given channel.
    frequency = QuantityDictFeat(None, 'L{key}F{value}', keys=CHANNELS, units='Hz')

    #: Power for a given channel (in db).
    powerdb = QuantityDictFeat(None, 'L{key}D{value}', keys=CHANNELS)

    #: Power for a given channel (in digital units).
    power = QuantityDictFeat(None, 'L{key}P{value:04d}', keys=CHANNELS, limits=(0, 1023, 1))


if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test Kentech HRI')
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False, help='Show interactive GUI')
    parser.add_argument('-p', '--port', type=str, default='17',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)
    with MDSnC.from_serial_port(args.port) as inst:
        if args.interactive:
            from lantz.ui.app import start_test_app
            start_test_app(inst)
        else:
            from time import sleep
            print("init")
            freq = 130
            inst.power(4,10)
            sleep(0.2)
            inst.enabled(4,1)
            sleep(1.2)
            inst.enabled(4,0)
            sleep(1.2)
            inst.enabled(4,1)





# Lakeshore 332 Temperature Controller Driver
# Peter Mintun <pmintun@uchicago.edu>

# This file is a driver for the Lakeshore 332 series temperature controller.

# Some sort of license information goes here.

from lantz import Feat, Action, DictFeat
from lantz.messagebased import MessageBasedDriver
from time import sleep


class Lakeshore332(MessageBasedDriver):
    """
    Lakeshore 332 Temperature controlller.

    This class, based off of the Lantz MessageBasedDriver class, implements a
    set of basic controls for the Lakeshore 332 series temperature controller.
    It essentially a port of a nice driver written for QtLab by Reinier Heeres.

    Full documentation of the device is available at:
    http://www.lakeshore.com/ObsoleteAndResearchDocs/332_Manual.pdf
    """

    # These defaults assume that you have set the IEEE Term setting to: Lf Cr
    DEFAULTS = {'COMMON': {'write_termination': '\n',
                           'read_termination': ''}}

    GPIB_name = None
    GPIB_address = -1

    channels = ['a', 'b']
    loops = ['1', '2']
    heater_range_vals = {'off': 0, 'low': 1, 'medium': 2, 'high': 3}
    heater_status_vals = {'no error': 0, 'open load': 1, 'short': 2}
    controller_modes = {'local': 0, 'remote': 1, 'remote, local lockout': 2}
    cmodes = {'manual PID': 1, 'zone': 2, 'open loop': 3, 'AutoTune PID': 4,
              'AutoTune PI': 5, 'AutoTune P': 6}

    T_min = 0
    T_max = 350

    T_min_set = 1.8
    T_max_set = 350

    _verbose = True

    @Feat()
    def idn(self):
        """
        Returns the instrument identification.
        """
        print('getting IDN')
        return self.query('*IDN?')

    @Action()
    def reset(self):
        """
        Resets the Lakeshore 332 temperature controller.
        """
        self.write('*RST')

    @DictFeat(limits=(T_min, T_max), keys=channels)
    def kelvin_meas(self, channel):
        """
        Returns measured temperature reading from specified channel in Kelvin.
        """
        return float(self.query('KRDG?{}'.format(channel)))

    @DictFeat(keys=channels)
    def sensor(self, channel):
        """
        Returns sensor reading from specified channel.
        """
        return float(self.query('SRDG?{}'.format(channel)))

    @Feat(values=heater_status_vals)
    def heater_status(self):
        """
        Returns the heater status.
        """
        return int(self.query('HTRST?'))

    @Feat(values=heater_range_vals)
    def heater_range(self):
        """
        Queries the instrument, prints a message describing the current heater
        range setting, then returns the heater range value.
        """
        return int(self.query('RANGE?'))

    @heater_range.setter
    def heater_range(self, heater_setting):
        """
        Sets heater range to  heater_setting.

        heater_setting must be an integer between 0 and 3 inclusive.
        """
        return self.write('RANGE {}'.format(heater_setting))

    @Feat()
    def heater_output_1(self):
        """
        Returns Loop 1 heater output in percent (%).
        """
        return float(self.query('HTR?'))

    @Feat()
    def heater_output_2(self):
        """
        Returns Loop 2 heater output in percent (%).
        """
        return float(self.query('AOUT?'))

    @Feat(values=controller_modes)
    def mode(self):
        """
        Reads the mode setting of the controller.
        """
        return int(self.query('MODE?'))

    @mode.setter
    def mode(self, mode):
        """
        Sets controller mode, valid mode inputs are:
        local (0)
        remote (1)
        remote, local lockout (2)
        """
        return self.query('MODE{}'.format(mode))

    @DictFeat(keys=loops)
    def pid(self, loop):
        """
        Get parameters for PID loop.
        """
        return self.query('PID?{}'.format(loop))

    @pid.setter
    def pid(self, loop, pid):
        """
        Get parameters for PID loop
        """
        p = pid[0]
        i = pid[1]
        d = pid[2]
        return self.query('PID{},{},{},{}'.format(loop, p, i, d))

    @DictFeat(limits=(T_min_set, T_max_set), keys=loops)
    def setpoint(self, channel):
        """
        Return the temperature controller setpoint.
        """
        return float(self.query('SETP?{}'.format(channel)))

    @setpoint.setter
    def setpoint(self, loop, T_set):
        """
        Sets the setpoint of channel channel to value value
        """
        self.query('SETP{},{}'.format(loop, T_set))
        sleep(0.05)
        return

    @DictFeat(limits=(0, 100), keys=loops)
    def mout(self, loop):
        """
        Returns loop manual heater power output.
        """
        return self.query('MOUT?{}'.format(loop))

    @mout.setter
    def mout(self, loop, percent):
        """
        Sets loop manual heater power output in percent.
        """
        return self.query('MOUT{},{}'.format(loop, percent))

    @DictFeat(values=cmodes, keys=loops)
    def cmode(self, loop):
        """
        Returns the control mode according to the following table.
        'manual PID' (1)
        'zone'(2)
        'open loop' (3)
        'AutoTune PID' (4)
        'AutoTune PI' (5)
        'AutoTune P' (6)
        """
        return int(self.query('CMODE?{}'.format(loop)))

    @cmode.setter
    def cmode(self, loop, value):
        """
        Sets the control mode according to the following table.
        'manual PID' (1)
        'zone'(2)
        'open loop' (3)
        'AutoTune PID' (4)
        'AutoTune PI' (5)
        'AutoTune P' (6)
        """
        return self.query('CMODE{},{}'.format(loop, value))


if __name__ == '__main__':
    with Lakeshore332('GPIB0::16::INSTR') as inst:
        print('The instrument identification is ' + inst.idn)

        print('resetting...')
        inst.reset
        print('reset.')

        # Testing mode switching functionality
        print('The current mode is ' + inst.mode + '.')
        inst.mode = 'remote, local lockout'
        print('Now the mode is ' + inst.mode + '.')
        inst.mode = 'remote'
        print('Now the mode is ' + inst.mode + '.')

        # Testing Kelvin read functionality
        print('Current temperature on channel a is ' +
              str(inst.kelvin_meas['a']) + ' Kelvin')
        print('Current temperature on channel b is ' +
              str(inst.kelvin_meas['b']) + ' Kelvin')

        # Testing sensor reading functionality
        print('Sensor reading on channel a is ' + str(inst.sensor['a']))
        print('Sensor reading on channel b is ' + str(inst.sensor['b']))

        # Testing heater status
        print('Heater status is ' + str(inst.heater_status))

        # Testing heater range
        print('Heater range is ' + str(inst.heater_range))
        inst.heater_range = 'low'
        print('Heater range is ' + str(inst.heater_range))
        inst.heater_range = 'off'
        print('Heater range is ' + str(inst.heater_range))

        # Testing heater output
        print('Loop 1 heater output ' + str(inst.heater_output_1) + '%')
        print('Loop 2 heater output ' + str(inst.heater_output_2) + '%')

        # Testing manual output
        print('Loop 1 manual output ' + str(inst.mout['1']))
        print('Loop 2 manual output ' + str(inst.mout['2']))
        inst.mout['1'] = 50
        inst.mout['2'] = 50
        print('Loop 1 manual output ' + str(inst.mout['1']))
        print('Loop 2 manual output ' + str(inst.mout['2']))
        inst.mout['1'] = 0
        inst.mout['2'] = 0
        print('Loop 1 manual output ' + str(inst.mout['1']))
        print('Loop 2 manual output ' + str(inst.mout['2']))

        # Testing cmode
        print('Loop 1 Command Mode: ' + str(inst.cmode['1']))
        inst.cmode['1'] = 'open loop'
        print('Loop 1 Command Mode: ' + str(inst.cmode['1']))
        inst.cmode['1'] = 'AutoTune P'
        print('Loop 1 Command Mode: ' + str(inst.cmode['1']))
        print('Loop 2 Command Mode: ' + str(inst.cmode['2']))

        # Testing setpoint
        inst.setpoint['1'] = 25
        print('Loop 1 setpoint is ' + str(inst.setpoint['1']))
        inst.setpoint['1'] = 50
        print('Loop 1 setpoint is ' + str(inst.setpoint['1']))
        inst.setpoint['1'] = 300
        print('Loop 1 setpoint is ' + str(inst.setpoint['1']))
        inst.setpoint['2'] = 300
        print('Loop 2 setpoint is ' + str(inst.setpoint['2']))

        # Testing PID
        inst.pid['1'] = list([10.0, 10.0, 10.0])
        print('Loop 1 PID parameters:' + str(inst.pid['1']))
        inst.pid['1'] = list([50.0, 20.0, 1.0])
        print('Loop 1 PID parameters:' + str(inst.pid['1']))
        print('Loop 2 PID parameters:' + str(inst.pid['2']))



# -*- coding: utf-8 -*-
"""
    dummydrivers
    ~~~~~~~~~~~~

    Just some fake drivers to show how the backend, frontend works.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import numpy as np

from lantz import Driver, Feat, ureg, Action


class DummyFunGen(Driver):
    """A Function Generator Driver.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._amplitude = ureg.volt
        self._frequency = ureg.Hz

    @Feat(units='Hz')
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, value):
        self._frequency = value

    @Feat(units='volt')
    def amplitude(self):
        return self._amplitude

    @amplitude.setter
    def amplitude(self, value):
        self._amplitude = value


class DummyOsci(Driver):
    """An Oscilloscope Driver.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._amplitude = ureg.volt

    @Action()
    def measure(self):
        return np.random.random((100, ))


class DummyShutter(Driver):
    """A Shutter Driver.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._opened = False

    @Feat(values={True, False})
    def opened(self):
        return self._opened

    @opened.setter
    def opened(self, value):
        self._opened = value



# -*- coding: utf-8 -*-
from lantz.messagebased import MessageBasedDriver
from lantz import Feat
import time
#from lantz.drivers.coherent.fieldmastergs import FieldMasterGS

class PilotBox(MessageBasedDriver):

    DEFAULTS = {
        'COMMON': {
            'read_termination': '\r\n',
            'write_termination': '\r',
        },
    }

    # def initialize(self):
    #     print("turning off this goddamn echo shit...")
    #     self.write('system:echo off')


    @Feat(read_once=True)
    def idn(self):
        return self.query('*idn?')

    @Feat(read_once=True)
    def echo(self):
        return self.query('system:echo off')

    @Feat(read_once=True)
    def min(self):
        print(self.query(':Laser:ILIMit? MIN'))
        return float(self.query(':Laser:ILIMit? MIN'))


    @Feat(read_once=True)
    def max(self):
        return float(self.query(':Laser:ILIMit? MAX'))


    @Feat(read_once=True, values={'Current': 'I', 'Power': 'P'})
    def laser_mode(self):
        mode = self.query(':Laser:MODe?')
        return mode[0]
    @laser_mode.setter
    def laser_mode(self, mode):
        self.write(':Laser:MODe {}'.format(mode))


    @Feat(read_once=True)
    def current_meas(self):
        return self.query(':Laser:CURRent?')


    @Feat(read_once=True)
    def current_set(self):
        return self.query(':Laser:CURRent:Set?')
    @current_set.setter
    def current_set(self, current):
        self.write(':Laser:CURRent {}A'.format(current))


    @Feat(read_once=True)
    def laser_status(self):
        return self.query(':Laser:STATus?')
    @laser_status.setter
    def laser_status(self, status):
        self.write(':Laser:STATus {}'.format(status))

    @Feat(read_once=True)
    def piezo_mod(self):
        return self.query(':Piezo:FREQuency:GENerator ?')

    @Feat(read_once=True)
    def piezo_amp(self):
        return self.query(':Piezo:FREQuency:ampl ?')

    @Feat(read_once=True, values={False:'OFF', True:'ON'})
    def piezo_status(self):
        return self.query(':Piezo:ENAble ?')

    @Feat(read_once=True)
    def piezo_offset(self):
        return self.query(':Piezo:OFFset ?')

    @Feat(read_once=True)
    def piezo_freq(self):
        return self.query(':Piezo:FREQuency ?')

    @piezo_freq.setter
    def piezo_freq(self,freq):
        self.write(':Piezo:FREQuency {}'.format(freq))

    @piezo_offset.setter
    def piezo_offset(self,offset):
        self.write(':Piezo:OFFset {}V'.format(offset))

    @piezo_status.setter
    def piezo_status(self,status,values={False:'OFF', True:'ON'}):
        self.write(':Piezo:ENAble: {}')


    @piezo_amp.setter
    def piezo_amp(self,ampl):
        self.write(':Piezo:FREQuency:ampl {}V'.format(ampl))

    @piezo_mod.setter
    def piezo_mod(self,mod):
        self.write(':Piezo:FREQuency:GENerator {}'.format(mod))
        #off, sin, tri

    @Feat(read_once=True)
    def current_lim(self):
        return self.query(':Laser:ILIMit? MAX')

    @current_lim.setter
    def current_lim(self,currentlim):
        self.write(':Laser:ILIMit {}A'.format(currentlim))




import time
if __name__ == '__main__':
    s = PilotBox('COM9')
    with s as pbox:
        #pbox.initialize()
        print(pbox.idn)
        print(pbox.echo)
        print(pbox.min)
        print(pbox.max)
        #print(pbox.laser_mode)
        #pbox.laser_mode = 'Current'
        #print(pbox.laser_mode)
        #print(pbox.current_meas)
        #print(pbox.current_set)
        #print(pbox.current_meas)
        #print(pbox.laser_status)
        #pbox.laser_status = 'ON'
        #time.sleep(5)
        #print(pbox.laser_status)
        #print(pbox.current_meas)
        #pbox.laser_status = 'OFF'
        #time.sleep(1)
        print(pbox.laser_status)
        print(pbox.current_set)
        print(pbox.current_meas)



"""
For some reason the actual measured laser current is always zero
Maybe the detector is broken or something
"""



#



from lantz import Feat, DictFeat, Action
from lantz.messagebased import MessageBasedDriver

from pyvisa import constants

from numpy import abs, ceil

from time import sleep


class SP2150i(MessageBasedDriver):
    """
    Implements controls for the Princeton Instruments Acton Series SP2150i
    Monochromater over the internal USB virtual serial port.

    Communication with the device is a little finnicky, so if you run into
    problems, I would suggest adding more buffer clears to avoid garbage
    accumulating in the buffer and messing up your commands.

    Author: Peter Mintun (pmintun@uchicago.edu)
    Date: 1/21/2016
    """
    DEFAULTS = {'ASRL': {'write_termination': '\r',
                         'read_termination': 'ok\r\n',
                         'baud_rate': 9600,
                         'data_bits': 8,
                         'parity': constants.Parity.none,
                         'stop_bits': constants.StopBits.one,
                         'encoding': 'latin-1',
                         'timeout': 2000}}

    max_speed = 100
    wavelength_min = 380
    wavelength_max = 520

    num_gratings = 8

    def initialize(self):
        """
        """
        super().initialize()
        self.clear_buffer()

    def clear_buffer(self):
        """
        This function sends an empty query just to clear any junk from the read
        buffer...This could probably be done more elegantly...but it works, for
        now at least.
        """
        return self.query('')

    @Feat(limits=(wavelength_min, wavelength_max))
    def nm(self):
        """
        Returns current wavelength of monochromater.
        """
        self.clear_buffer()
        read = self.query('?NM')
        wavelength = read.replace('nm', '')
        return float(wavelength)

    @nm.setter
    def nm(self, wavelength):
        """
        Sets output to specified wavelength, traveling at the current scan
        rate.
        """
        curr_wavelength = self.nm
        scan_rate = self.scan_speed
        delta_lambda = abs(curr_wavelength - wavelength)
        if delta_lambda > 0.1:
            scan_time = ceil(delta_lambda / scan_rate * 60) + 2  # need seconds
            self.clear_buffer()
            print('Scanning from {}nm to {}nm, will take {}sec'.format(
                curr_wavelength, wavelength, scan_time))
            return self.write_message_wait('{0:.1f} NM'.format(wavelength),
                                           scan_time)
        else:
            return


    @Feat(limits=(0, max_speed))
    def scan_speed(self):
        """
        Get scan rate in nm/min.
        """
        self.clear_buffer()
        read = self.query('?NM/MIN')
        speed = read.replace('nm/min', '')
        return float(speed)

    @scan_speed.setter
    def scan_speed(self, speed):
        """
        Sets current scan speed in nm/min.
        """
        return self.write_message_wait('{0:.1f} NM/MIN'.format(speed), 0.1)

    @Feat(limits=(1, 2, 1))
    def grating(self):
        """
        Returns the current grating position
        """
        response = self.query('?GRATING')
        return int(response)

    @grating.setter
    def grating(self, grating_num):
        """
        Changes the current grating to be the one in slot grating_num.
        """
        print('Warning: Untested code! No other gratings were installed.')
        print('Changing grating, please wait 20 seconds...')
        return self.write_message_wait('{} GRATING'.format(grating_num), 20)

    @Feat(limits=(1, 3, 1))
    def turret(self):
        """
        Returns the selected turret number.
        """
        response = self.query('?TURRET')
        return int(response.replace('  ok', ''))

    @turret.setter
    def turret(self, turr_set):
        """
        Selects the parameters for the grating on turret turr_set
        """
        print('Warning: untested. No other turrets were installed.')
        return self.write_message_wait('{} TURRET'.format(turr_set), 1)

    @Feat()
    def turret_spacing(self):
        """
        Returns the groove spacing of the grating for each turret.
        """
        print('Warning: This command does\'t do anything?')
        return self.write_message_wait('?TURRETS', 0.5)

    @Feat()
    def grating_settings(self):
        """
        Returns the groove spacing and blaze wavelength of grating positions
        for all possible gratings.
        """
        gratings = []
        self.write('?GRATINGS')
        self.read()
        for i in range(0, self.num_gratings):
            gratings.append(self.read())
        self.read()
        return gratings

    def write_message_wait(self, message, wait_time):
        self.write(message)
        sleep(wait_time)
        return self.read()

if __name__ == '__main__':
    with SP2150i('ASRL4::INSTR') as inst:
        print('== Monochromater Wavelength ==')
        print('Wavelength:{}nm'.format(inst.nm))
        print('Scan rate:{}nm/min'.format(inst.scan_speed))

        print('== Monochromater Information ==')
        print('Selected turret:{}'.format(inst.turret))
        print('Selected grating:{}'.format(inst.grating))
        # inst.turret = 3
        # inst.grating = 2
        print('Selected turret:{}'.format(inst.turret))
        print('Selected grating:{}'.format(inst.grating))
        print('Turret Spacing:{}'.format(inst.turret_spacing))
        print('Gratings:{}'.format(inst.grating_settings))

        for i in range(1, 20):
            inst.scan_speed = 100.0
            print('Scan rate:{}nm/min'.format(inst.scan_speed))
            print('Wavelength:{}nm'.format(inst.nm))
            inst.nm = 400.0
            print('Wavelength:{}nm'.format(inst.nm))
            inst.scan_speed = 50.0
            print('Scan rate:{}nm/min'.format(inst.scan_speed))
            inst.nm = 500.0
            print('Wavelength:{}nm'.format(inst.nm))
            print('Cycle {} complete.'.format(i))



"""
    lantz.drivers.tektronix.awg5000
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Drivers for the Tektronix 5000 series AWG

    Authors: Alexandre Bourassa, Kevin Miao
    Date: September 14, 2016
"""

import numpy as np
import re
from enum import Enum
import ftplib as _ftp
import os as _os
import time as _t

from lantz.messagebased import MessageBasedDriver
from lantz import Feat, DictFeat, Action


class AWGState(Enum):
    stopped = 0
    trigger_wait = 1
    running = 2


_ch_markers = [(ch, m) for ch in range(1, 5) for m in range(1, 3)]

class AWG5000(MessageBasedDriver):

    DEFAULTS = {
        'COMMON': {
            'write_termination': '\n',
            'read_termination': '\n',
            'timeout': 10000,
        },
    }

    def __init__(self, resource_name, ftp_ip=None, *args, **kwargs):
        """If ftp_ip is None, it will attempt to figure out the ftp ip from the resources name
        """
        super().__init__(resource_name, *args, **kwargs)
        if ftp_ip is None:
            self.ip = re.findall('[0-9]+.[0-9]+.[0-9]+.[0-9]+', resource_name)
        else:
            self.ip = ftp_ip
        return

    def initialize(self):
        super().initialize()

        #Start from the ftp directory
        self.cd('\\')
        self.cd('ftp')

    @Feat(read_once=True)
    def idn(self):
        return self.query('*IDN?')

    @Feat(values={item.name: item.value for item in AWGState})
    # @Feat()
    def toggle_run(self):
        # return AWGState(int(self.query('AWGC:RST?')))
        return int(self.query('AWGC:RST?'))

    @toggle_run.setter
    def toggle_run(self, state):
        if state or state == AWGState.running:
            cmd = 'RUN'
        elif not state or state == AWGState.stopped:
            cmd = 'STOP'
        else:
            raise ValueError('invalid run state: {}'.format(state))
        self.write('AWGC:{}:IMM'.format(cmd))

    @DictFeat(keys=range(1, 5), values={True: '1', False: '0'})
    def toggle_output(self, channel):
        return self.query('OUTP{}:STAT?'.format(channel))

    @toggle_output.setter
    def toggle_output(self, channel, state):
        self.write('OUTP{}:STAT {}'.format(channel, state))

    @Action()
    def toggle_all_outputs(self, state):
        for channel in range(1, 5):
            self.toggle_output[channel] = state

    @DictFeat(units='V', keys=_ch_markers)
    def marker_amplitude(self, ch_m):
        ch, m = ch_m
        return float(self.query('SOUR{}:MARK{}:VOLT:AMPL?'.format(ch, m)))

    @marker_amplitude.setter
    def marker_amplitude(self, ch_m, value):
        ch, m = ch_m
        self.write('SOUR{}:MARK{}:VOLT:AMPL {}V'.format(ch, m, value))
        return

    @DictFeat(units='V', keys=_ch_markers)
    def marker_high(self, ch_m):
        ch, m = ch_m
        return float(self.query('SOUR{}:MARK{}:VOLT:HIGH?'.format(ch, m)))

    @marker_high.setter
    def marker_high(self, ch_m, value):
        ch, m = ch_m
        self.write('SOUR{}:MARK{}:VOLT:HIGH {}V'.format(ch, m, value))
        return

    @DictFeat(units='V', keys=_ch_markers)
    def marker_low(self, ch_m):
        ch, m = ch_m
        return float(self.query('SOUR{}:MARK{}:VOLT:LOW?'.format(ch, m)))

    @marker_low.setter
    def marker_low(self, ch_m, value):
        ch, m = ch_m
        self.write('SOUR{}:MARK{}:VOLT:LOW {}V'.format(ch, m, value))
        return

    @DictFeat(units='V', keys=_ch_markers)
    def marker_offset(self, ch_m):
        ch, m = ch_m
        return float(self.query('SOUR{}:MARK{}:VOLT:OFFS?'.format(ch, m)))

    @marker_offset.setter
    def marker_offset(self, ch_m, value):
        ch, m = ch_m
        self.write('SOUR{}:MARK{}:VOLT:OFFS {}V'.format(ch, m, value))
        return

    @DictFeat(limits=(1, 2 ** 16))
    def seq_loop_count(self, line):
        return int(self.query('SEQUENCE:ELEMENT{}:LOOP:COUNT?'.format(line)))

    @seq_loop_count.setter
    def seq_loop_count(self, line, count):
        self.write('SEQUENCE:ELEMENT{}:LOOP:COUNT {}'.format(line, count))

    @DictFeat(values={True: '1', False: '0'})
    def seq_loop_infinite(self, line):
        return self.query('SEQUENCE:ELEMENT{}:LOOP:INFINITE?'.format(line))

    @seq_loop_infinite.setter
    def seq_loop_infinite(self, line, infinite_loop):
        self.write('SEQUENCE:ELEMENT{}:LOOP:INFINITE {}'.format(line, infinite_loop))

    @Action()
    def jump_to_line(self, line):
        self.write('SEQ:JUMP:IMM {}'.format(line))

    @Action()
    def trigger(self):
        self.write('*TRG')

    @DictFeat()
    def waveform(self, key):
        index, channel = key
        return self.query('SEQ:ELEM{}:WAV{}?'.format(index, channel))

    @waveform.setter
    def waveform(self, key, value):
        index, channel = key
        self.write('SEQ:ELEM{}:WAV{} {}'.format(index, channel, value))


    # ----------------------------------------------------
    # Hard Drive navigation method
    # ----------------------------------------------------
    VALID_DRIVE = {"MAIN", "FLOP", "NET1", "NET2", "NET3"}

    @Action()
    def cd(self, dir):
        self.write('MMEM:CDIR "{}"'.format(dir))

    @Action()
    def ls(self, verbose=True):
        #Query the current directory and the content
        dir = self.query("MMEM:CDIR?").strip('"')
        content = self.query("MMEM:CAT?").split(',"')
        used, avail = map(int, content.pop(0).split(","))

        dirs = list()
        files = list()
        for item in content:
            item = item.strip('"')
            name, isDir, size = item.split(',')
            if isDir=='DIR': dirs.append(name)
            else           : files.append((name, size))

        # Build and print an answer (or return it
        if verbose:
            spaces_size = 30
            ans =  "{} ({:.2f}% full):\r\n".format(dir, (used / (used + avail)))
            for d in dirs: ans += '\t|_ '+d+(' '*(spaces_size-len(d)))+'DIR\n'
            for f in files: ans += '\t|_ '+f[0]+(' '*(spaces_size-len(f[0])))+f[1]+'\n'
            print(ans)
        else:
            return (dir, used, avail), dirs, files

    @Action()
    def mkdir(self, dir_name, drive="MAIN"):
        if not drive in self.VALID_DRIVE: raise Exception("Invalid drive!")
        self.write(':MMEM:MDIR "{}", "{}"'.format(dir_name, drive))

    @Action()
    def select_drive(self, drive):
        if not drive in self.VALID_DRIVE: raise Exception("Invalid drive!")
        self.write(':MMEM:MSIS "{}"'.format(drive))

    @Action()
    def get_current_drive(self):
        print(self.query(':MMEM:MSIS?'))

    #This might not work
    @Action()
    def mv(self, source_file_path, dest_file_path, source_drive="MAIN", dest_drive="MAIN"):
        if not source_drive in self.VALID_DRIVE: raise Exception("Invalid source drive!")
        if not dest_drive in self.VALID_DRIVE: raise Exception("Invalid destination drive!")
        self.write(':MMEM:MOVE "{}","{}","{}","{}"'.format(source_file_path, source_drive, dest_file_path, dest_drive))

    # This might not work
    @Action()
    def cp(self, source_file_path, dest_file_path, source_drive="MAIN", dest_drive="MAIN"):
        if not source_drive in self.VALID_DRIVE: raise Exception("Invalid source drive!")
        if not dest_drive in self.VALID_DRIVE: raise Exception("Invalid destination drive!")
        self.write(':MMEM:COPY "{}","{}","{}","{}"'.format(source_file_path, source_drive, dest_file_path, dest_drive))

    def upload_file(self, local_filename, remote_filename, print_progress=True):
        self.ftp = _ftp.FTP(self.ip)
        self.ftp.login()
        if print_progress:
            FTP_Upload(self.ftp,local_filename, remote_filename)
        else:
            self.ftp.storbinary('STOR ' + remote_filename, open(local_filename, 'rb'), blocksize=1024)
        self.ftp.quit()

    @Action()
    def load_awg_file(self, filename):
        self.write('AWGCONTROL:SRESTORE "{}"'.format(filename))



class FTP_Upload():
    def __init__(self, ftp, local_filename, remote_filename):

        block_size = 1024
        total_size = _os.path.getsize(local_filename)
        self.written_size = 0
        self.last_time = _t.time()
        self.last_percent = 0

        def callback(block):
            self.written_size += block_size
            time = _t.time()
            percent = (self.written_size / total_size)*100
            #print(time - self.last_time > 1,percent - self.last_percent  > 0.1)
            if time - self.last_time > 1 and percent - self.last_percent  > 0.1:
                self.last_time = time
                self.last_percent = percent
                print('{:.2f}%'.format(percent))
        print('Uploading "{}" to remote destination "{}"'.format(local_filename, remote_filename))
        ftp.storbinary('STOR ' + remote_filename, open(local_filename, 'rb'), blocksize=block_size, callback=callback)


if __name__ == '__main__':
    test()



# winspec.py, COM wrapper for use with Roper Scientific's Winspec
# Reinier Heeres <reinier@heeres.eu>, 2009
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from comtypes.client import CreateObject, Constants
import numpy as np
import time

_exp = None
_app = None
_const = None
_spec_mgr = None
_spec = None

def _initialize():
    '''
    Create objects. Associated constants will be available in the _const
    object. Note that it's not possible to request the contents of that;
    one has to look in the Python file generated with gen_py.
    '''
    global _exp, _app, _const, _spec_mgr, _spec
    _exp = CreateObject('WinX32.ExpSetup.2')
    _app = CreateObject('WinX32.Winx32App.2')
    _const = Constants(_exp)
    _spec_mgr = CreateObject('WinX32.SpectroObjMgr')
    _spec = _spec_mgr.Current
    _spec.Process(_const.SPTP_INST_LOADCONFIGURATION)

def _get_wavelen(coord, coeffs):
    val = 0
    for power, coef in enumerate(coeffs):
        val += coef * (coord + 1) ** power
    return val

def get_exposure_time():
    return _exp.SGetParam(_const.EXP_EXPOSURE)[1]

def set_exposure_time(val):
    return _exp.SetParam(_const.EXP_EXPOSURE, val)

def get_use_new_window():
    return _exp.SGetParam(_const.EXP_NEWWINDOW)[1]

def set_use_new_window(val):
    return _exp.SetParam(_const.EXP_NEWWINDOW, val)

def get_file_inc_enable():
    return _exp.SGetParam(_const.EXP_FILEINCENABLE)[1]

def set_file_inc_enable(val):
    return _exp.SetParam(_const.EXP_FILEINCENABLE, val)

def get_autosave():
    return _exp.SGetParam(_const.EXP_AUTOSAVE)[1]

def set_autosave(val):
    return _exp.SetParam(_const.EXP_AUTOSAVE, val)

def get_grating():
    return _spec.GetParam(_const.SPT_CUR_GRATING)[1]

def get_grating_grooves(gr):
    '''gr is the absolute grating number.'''
    return _spec.GetParam(_const.SPT_GRAT_GROOVES, gr)[1]

def get_grating_name(gr):
    '''gr is the absolute grating number.'''
    return _spec.GetParam(_const.SPT_GRAT_USERNAME, gr)[1]

def get_ngratings():
    '''Get number of gratings per turret.'''
    return _spec.GetParam(_const.SPT_GRATINGSPERTURRET)[1]

def get_current_turret():
    return _spec.GetParam(_const.SPT_ACTIVE_TURRET_NUM)[1]

def set_grating(val):
    _spec.SetParam(_const.SPT_NEW_GRATING, val)
    _spec.Move()
    return get_grating()

def get_temperature():
    return _exp.SGetParam(_const.EXP_ACTUAL_TEMP)[1]

def get_target_temperature():
    return _exp.SGetParam(_const.EXP_TEMPERATURE)[1]

def set_target_temperature(val):
    return _exp.SetParam(_const.EXP_TEMPERATURE, val)

def get_wavelength():
    return _spec.GetParam(_const.SPT_CUR_POSITION)[1]

def set_wavelength(val):
    _spec.SetParam(_const.SPT_NEW_POSITION, float(val))
    _spec.Move()
    return get_wavelength()

# Default maximum number of sleeps for 250sec
MAX_SLEEPS = 5000
doc = None

def get_wavelengths():
    global doc
    if doc is None:
        doc = CreateObject('WinX32.DocFile.3')

    _exp.Start(doc)
    i = 0
    while i < MAX_SLEEPS:
        status = _exp.SGetParam(_const.EXP_RUNNING)[1]
        if status == 0:
            break
        i += 1
        time.sleep(0.05)

    if i == MAX_SLEEPS:
        print 'Warning: maximum delay exceeded'
        return None
    xdim = doc.SGetParam(_const.DM_XDIM)[1]
    ydim = doc.SGetParam(_const.DM_YDIM)[1]
    if ydim != 1:
        raise ValueError('Can only get 1D spectra')
    calib = doc.GetCalibration()
    order = calib.Order
    coeffs = [calib.PolyCoeffs(i) for i in range(order + 1)]
    wlens = [_get_wavelen(i, coeffs) for i in range(xdim)]
    return wlens


#def get_spectrum(wlen=True, wlenpoly=True, newdoc=False):
def get_spectrum(wlen=False, wlenpoly=True, newdoc=False):
    '''
    Get a spectrum using winspec.

    If wlen is True it returns a 2D numpy array with wavelength and counts
    as the columns. If wlen is False it returns a 1D array with counts.
    If wlenpoly is True the polynomial approximation for determining the
    wavelength will be used, which is quite a bit faster than querying winspec
    for the individual pixels. It can be off by ~0.3nm.
    '''

    global doc
    if newdoc or doc is None:
        doc = CreateObject('WinX32.DocFile.3')

    _exp.Start(doc)
    i = 0
    while i < MAX_SLEEPS:
        status = _exp.SGetParam(_const.EXP_RUNNING)[1]
        if status == 0:
            break
        i += 1
        time.sleep(0.05)

    if i == MAX_SLEEPS:
        print 'Warning: maximum delay exceeded'
        return None

    xdim = doc.SGetParam(_const.DM_XDIM)[1]
    ydim = doc.SGetParam(_const.DM_YDIM)[1]
    if ydim != 1:
        raise ValueError('Can only get 1D spectra')

    spectrum = np.array(doc.GetFrame(1),dtype=np.uint16).flatten()
    #spectrum = np.array(doc.GetFrame(1),dtype=np.uint16)
    #spectrum.flatten()
    if wlen:
        calib = doc.GetCalibration()
        if wlenpoly:
            order = calib.Order
            coeffs = [calib.PolyCoeffs(i) for i in range(order + 1)]
            wlens = [_get_wavelen(i, coeffs) for i in range(xdim)]
        else:
            wlens = [calib.Lambda(i + 1) for i in range(xdim)]
        spectrum = zip(wlens, spectrum)

    return spectrum

_initialize()




# -*- coding: utf-8 -*-
"""
    lantz.drivers.newport.motionesp301
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers to control ESP300 and ESP301 motion controller 
    via USB or serial.

    For USB, one first have to install the windows driver from newport.

    :copyright: 2015, see AUTHORS for more details.
    :license: GPL, see LICENSE for more details.
    
    Source: Instruction Manual (Newport)
"""


from lantz.feat import Feat
from lantz.action import Action
#from lantz.serial import SerialDriver
from lantz.messagebased import MessageBasedDriver
from pyvisa import constants
#from lantz.visa import GPIBVisaDriver
from lantz import Q_, ureg
from lantz.processors import convert_to
import time
import numpy as np
import copy

# Add generic units:
#ureg.define('unit = unit')
#ureg.define('encodercount = count')
#ureg.define('motorstep = step')


class ESP301(MessageBasedDriver):
    """ Newport ESP301 motion controller. It assumes all axes to have units mm

    :param scan_axes: Should one detect and add axes to the controller
    """

    DEFAULTS = {
                'COMMON': {'write_termination': '\r\n',
                    'read_termination': '\r\n',},
                'ASRL':{
                    'timeout': 4000, #ms
                    'encoding': 'ascii',
                    'data_bits': 8,
                    'baud_rate': 19200,
                    'parity': constants.Parity.none,
                    'stop_bits': constants.StopBits.one,
                    'flow_control': constants.VI_ASRL_FLOW_RTS_CTS,#constants.VI_ASRL_FLOW_NONE,
                    },
                }

    def initialize(self):
        super().initialize()
        self.detect_axis()

    @classmethod
    def via_usb(cls, port, name=None, **kwargs):
        """Connect to the ESP301 via USB. Internally this goes via serial"""
        cls.DEFAULTS = copy.deepcopy(cls.DEFAULTS)
        cls.DEFAULTS['ASRL'].update({'baud_rate': 921600})
        return cls.via_serial(port=port, name=name, **kwargs)


    @Action()
    def detect_axis(self):
        """ Find the number of axis available.
        
        The detection stops as soon as an empty controller is found.
        """
        self.axes = []
        i = 0
        scan_axes = True
        while scan_axes:
            try:
                i += 1
                id = self.query('%dID?' % i)
                axis = ESP301Axis(self, i, id)
                self.axes.append(axis)
            except:
                err = self.get_errors()
                if err == 37:  # Axis number missing
                    self.axes.append(None)
                elif err == 9:  # Axis number out of range
                    scan_axes = False
                elif err == 6:  # Axis number out of range, but wrong errorcode
                    scan_axes = False
                else:  # Dunno...
                    raise Exception(err)
    @Action()
    def get_errors(self):
        err = int(self.query('TE?'))
        return err
 
    @Feat(read_once=False)
    def position(self):
        return [axis.position for axis in self.axes]

    @Feat(read_once=False)
    def _position_cached(self):
        return [axis._position_cached for axis in self.axes]

    @position.setter
    def position(self, pos):
        """Move to position (x,y,...)"""
        return self._position(pos)

    @Action()
    def _position(self, pos, read_pos=None, wait_until_done=True):
        """Move to position (x,y,...)"""
        if read_pos is not None:
            self.log_error('kwargs read_pos for function _position is deprecated')

        for p, axis in zip(pos, self.axes):
            if not p is None:
                axis._set_position(p, wait=False)
        if wait_until_done:
            for p, axis in zip(pos, self.axes):
                if not p is None:
                    axis._wait_until_done()
                    axis.check_position(p)
            return self.position

        return pos
    
    @Action()
    def motion_done(self):
        for axis in self.axes: axis._wait_until_done()
                
    def finalize(self):
        for axis in self.axes:
            if axis is not None:
                del (axis)
        super().finalize()




#class ESP301GPIB( ESP301, GPIBVisaDriver):
#    """ Untested!
#    """
#    def __init__(self, scan_axes=True, resource_name= 'GPIB0::2::INSTR', *args, **kwargs):
#        # Read number of axes and add axis objects
#        self.scan_axes = scan_axes
#        super().__init__(resource_name=resource_name, *args, **kwargs)


class ESP301Axis(ESP301):
    def __init__(self, parent, num, id, *args, **kwargs):
        #super(ESP301Axis, self).__init__(*args, **kwargs)
        self.parent = parent
        self.num = num
        self.id = id
        self.wait_time = 0.01 # in seconds * Q_(1, 's')
        self.backlash = 0
        self.wait_until_done = True
        self.accuracy = 0.001 # in units reported by axis
        # Fill position cache:
        self.position

    def __del__(self):
        self.parent = None
        self.num = None

    def id(self):
        return self.id

    @Action()
    def on(self):
        """Put axis on"""
        self.parent.write('%dMO' % self.num)

    @Action()
    def off(self):
        """Put axis on"""
        self.parent.write('%dMF' % self.num)

    @Feat(values={True: '1', False: '0'})
    def is_on(self):
        """
        :return: True is axis on, else false
        """
        return self.parent.query('%dMO?' % self.num)

    @Action(units='mm')
    def define_home(self, val=0):
        """Remap current position to home (0), or to new position

        :param val: new position"""
        self.parent.write('%dDH%f' % (self.num, val))

    @Feat(units='mm', read_once=False)
    def _position_cached(self):
        return self.__position_cached
    
    @_position_cached.setter
    def _position_cached(self, pos):
        self.__position_cached = pos

    @Feat(units='mm')
    def position(self):
        self._position_cached = float(self.parent.query('%dTP?' % self.num))*ureg.mm
        return self._position_cached

    @position.setter
    def position(self, pos):
        """
        Waits until movement is done if self.wait_until_done = True.

        :param pos: new position
        """
        if not self.is_on:
            self.log_error('Axis not enabled. Not moving!')
            return

        # First do move to extra position if necessary
        self._set_position(pos, wait=self.wait_until_done)

    @Action(units=['mm',None])
    def _set_position(self, pos, wait=None):
        """
        Move to an absolute position, taking into account backlash.

        When self.backlash is to a negative value the stage will always move
         from low to high values. If necessary, a extra step with length
         self.backlash is set.

        :param pos: New position in mm
        :param wait: wait until stage is finished
        """

        # First do move to extra position if necessary
        if self.backlash:
            position = self.position.magnitude
            #backlash = self.backlash.to('mm').magnitude
            backlash = convert_to('mm', on_dimensionless='ignore')(self.backlash).magnitude
            if ( backlash < 0 and position > pos) or\
                ( backlash > 0 and position < pos):

                self.log_info('Using backlash')
                self.__set_position(pos + backlash)
                self._wait_until_done()

        # Than move to final position
        self.__set_position(pos)
        if wait:
            self._wait_until_done()
            self.check_position(pos)

    def __set_position(self, pos):
        """
        Move stage to a certain position
        :param pos: New position
        """
        self.parent.write('%dPA%f' % (self.num, pos))

    def check_position(self, pos):
        '''Check is stage is at expected position'''
        if np.isclose(self.position, pos, atol=self.accuracy):
            return True
        self.log_error('Position accuracy {} is not reached.'
            'Expected: {}, measured: {}'.format(self.accuracy, pos, self._position_cached))
        return False

    @Feat(units='mm/s')
    def max_velocity(self):
        return float(self.parent.query('%dVU?' % self.num))

    @max_velocity.setter
    def max_velocity(self, velocity):
        self.parent.write('%dVU%f' % (self.num, velocity))
    
    @Feat(units='mm/s**2')
    def max_acceleration(self):
        return float(self.parent.query('%dAU?' % self.num))

    @max_acceleration.setter
    def max_acceleration(self, velocity):
        self.parent.write('%dAU%f' % (self.num, velocity))


    @Feat(units='mm/s')
    def velocity(self):
        return float(self.parent.query('%dVA?' % self.num))

    @velocity.setter
    def velocity(self, velocity):
        """
        :param velocity: Set the velocity that the axis should use when moving
        :return:
        """
        self.parent.write('%dVA%f' % (self.num, velocity))
    
    @Feat(units='mm/s**2')
    def acceleration(self):
        return float(self.parent.query('%dVA?' % self.num))

    @acceleration.setter
    def acceleration(self, acceleration):
        """
        :param acceleration: Set the acceleration that the axis should use when starting
        :return:
        """
        self.parent.write('%dAC%f' % (self.num, acceleration))

    @Feat(units='mm/s')
    def actual_velocity(self):
        return float(self.parent.query('%dTV' % self.num))

    @actual_velocity.setter
    def actual_velocity(self, val):
        raise NotImplementedError

    @Action()
    def stop(self):
        """Emergency stop"""
        self.parent.write(u'{0:d}ST'.format(self.num))

    @Feat(values={True: '1', False: '0'})
    def motion_done(self):
        while True:
            ret = self.parent.query('%dMD?' % self.num)
            if ret in ['1','0']:
                break
            else:
                time.sleep(self.wait_time)
        return ret

    # Not working yet, see https://github.com/hgrecco/lantz/issues/35
    # @Feat(values={Q_('encodercount'): 0,
    #                     Q_('motor step'): 1,
    #                     Q_('millimeter'): 2,
    #                     Q_('micrometer'): 3,
    #                     Q_('inches'): 4,
    #                     Q_('milli-inches'): 5,
    #                     Q_('micro-inches'): 6,
    #                     Q_('degree'): 7,
    #                     Q_('gradian'): 8,
    #                     Q_('radian'): 9,
    #                     Q_('milliradian'): 10,
    #                     Q_('microradian'): 11})
    def units(self):
        ret =  int(self.parent.query(u'{}SN?'.format(self.num)))
        vals = {0 :'encoder count',
                 1 :'motor step',
                 2 :'millimeter',
                 3 :'micrometer',
                 4 :'inches',
                 5 :'milli-inches',
                 6 :'micro-inches',
                 7 :'degree',
                 8 :'gradian',
                 9 :'radian',
                 10:'milliradian',
                 11:'microradian',}
        return vals[ret]

    # @units.setter
    # def units(self, val):
    #     self.parent.write('%SN%' % (self.num, val))

    def _wait_until_done(self):
        #wait_time = convert_to('seconds', on_dimensionless='warn')(self.wait_time)
        time.sleep(self.wait_time)
        while not self.motion_done:
            time.sleep(self.wait_time) #wait_time.magnitude)



if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test ESP301 driver')
    parser.add_argument('-p', '--port', type=str, default='1',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantz.log.log_to_socket(lantz.log.DEBUG)

    with ESP301.via_usb(port=args.port) as inst:
        # inst.initialize() # Initialize the communication with the power meter
        # Find the status of all axes:
        for axis in inst.axes:
            print('Axis {} Position {} is_on {} max_velocity {} velocity {}'.format(axis.num, axis.position,
                                                                                    axis.is_on, axis.max_velocity,
                                                                                    axis.velocity))



from lantz import Feat
from lantz.messagebased import MessageBasedDriver

class Lakeshore211(MessageBasedDriver):

    DEFAULTS = {
        'COMMON': {
            'write_termination': '\r\n',
            'read_termination': '\r\n',
        }
    }

    @Feat(units='Kelvin')
    def temperature(self):
        return float(self.query('KRDG?'))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.scpi
    ~~~~~~~~~~~~~~~~~~

    Implements the SCPI protocol.

    Standard Commands for Programmable Instruments

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Action, Feat, Driver
from lantz.drivers.ieee4882 import IEEE4882Driver


def build_feat(command, **kwargs):
    """Builds a feat with with SCPI. Can be used as decorator
    for an empty function providing only the name and docstring.

    :param command: SCPI command

    Keyword arguments will be given to the Feat constructor.
    (e.g. units, limits, etc)

    Special behaviours for kwargs:
    - If the 'values' is a string, a set will be created by splitting on |


    Example:

        @build_feat('MEAS')
        def measure():
            "Meas docstring"

    If the command ends in ?, the Feat will be read only (no setter)
    If you want a write only Feat (no getter), append a ! to the command.
    (but consider if this is not actually an action)
    """

    if 'values' in kwargs and isinstance(kwargs['values'], str):
        kwargs['values'] = set(kwargs['values'].split('|'))

    def deco(func):

        nonlocal command
        if command.endswith('!'):
            command = command[:-1]
            getter = None
        else:
            def getter(self):
                return self.query(command + '?')

        if command.endswith('?'):
            setter = None
        else:
            def setter(self, value):
                return self.send(command + str(value))

        return Feat(getter, setter, doc=func.__doc__, **kwargs)

    return deco


class SCPIDriver(IEEE4882Driver):
    """Implements mandatory functions for a SCPI Device.

    You can use it as a mixin class.
    """



# -*- coding: utf-8 -*-
"""
    lantz.drivers.newport.motionsmc100
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Implements the drivers to control SMC100 controller

    :copyright: 2018, see AUTHORS for more details.
    :license: GPL, see LICENSE for more details.

    Source: Instruction Manual (Newport)

"""


from lantz.feat import Feat
from lantz.action import Action
import pyvisa
from pyvisa import constants

from lantz import Q_, ureg
from lantz.processors import convert_to
from lantz.messagebased import MessageBasedDriver

from lantz.drivers.newport_motion.motion import MotionAxis
from lantz.drivers.motion import MotionControllerMultiAxis

import time
import numpy as np

ERRORS = {"@": "",
          "A": "Unknown message code or floating point controller address.",
          "B": "Controller address not correct.",
          "C": "Parameter missing or out of range.",
          "D": "Execution not allowed.",
          "E": "home sequence already started.",
          "I": "Execution not allowed in CONFIGURATION state.",
          "J": "Execution not allowed in DISABLE state.",
          "H": "Execution not allowed in NOT REFERENCED state.",
          "K": "Execution not allowed in READY state.",
          "L": "Execution not allowed in HOMING state.",
          "M": "Execution not allowed in MOVING state.",
          }

positioner_errors = {
    0b1000000000: '80 W output power exceeded',
    0b0100000000: 'DC voltage too low',
    0b0010000000: 'Wrong ESP stage',
    0b0001000000: 'Homing time out',
    0b0000100000: 'Following error',
    0b0000010000: 'Short circuit detection',
    0b0000001000: 'RMS current limit',
    0b0000000100: 'Peak current limit',
    0b0000000010: 'Positive end of run',
    0b0000000001: 'Negative end of run',
                    }
controller_states = {
    '0A': 'NOT REFERENCED from reset.',
    '0B': 'NOT REFERENCED from HOMING.',
    '0C': 'NOT REFERENCED from CONFIGURATION.',
    '0D': 'NOT REFERENCED from DISABLE.',
    '0E': 'NOT REFERENCED from READY.',
    '0F': 'NOT REFERENCED from MOVING.',
    '10': 'NOT REFERENCED ESP stage error.',
    '11': 'NOT REFERENCED from JOGGING.',
    '14': 'CONFIGURATION.',
    '1E': 'HOMING commanded from RS-232-C.',
    '1F': 'HOMING commanded by SMC-RC.',
    '28': 'MOVING.',
    '32': 'READY from HOMING.',
    '33': 'READY from MOVING.',
    '34': 'READY from DISABLE.',
    '35': 'READY from JOGGING.',
    '3C': 'DISABLE from READY.',
    '3D': 'DISABLE from MOVING.',
    '3E': 'DISABLE from JOGGING.',
    '46': 'JOGGING from READY.',
    '47': 'JOGGING from DISABLE.',
                    }


class SMC100(MessageBasedDriver, MotionControllerMultiAxis):
    """ Newport SMC100 motion controller. It assumes all axes to have units mm


    Example:
    import numpy as np
    import lantz
    import visa
    import lantz.drivers.pi.piezo as pi
    from lantz.drivers.newport_motion import SMC100
    from pyvisa import constants
    rm = visa.ResourceManager('@py')
    lantz.messagebased._resource_manager = rm
    ureg = lantz.ureg
    try:
        lantzlog
    except NameError:
        lantzlog = lantz.log.log_to_screen(level=lantz.log.DEBUG)
        lantz.log.log_to_socket(level=lantz.log.DEBUG)
        
    import time
    import numpy as np
    import warnings
    #warnings.filterwarnings(action='ignore')
    print(lantz.messagebased._resource_manager.list_resources())
    stage = SMC100('ASRL/dev/ttyUSB0::INSTR')
    stage.initialize()
    axis0 = stage.axes[0]
    print('Axis id:' + axis0.idn)
    print('Axis position: {}'.format(axis0.position))
    axis0.keypad_disable()
    axis0.position += 0.1 * ureg.mm
    print('Errors: {}'.format(axis0.get_errors()))
    stage.finalize()
    """

    DEFAULTS = {
                'COMMON': {'write_termination': '\r\n',
                           'read_termination': '\r\n', },
                'ASRL': {
                    'timeout': 100,  # ms
                    'encoding': 'ascii',
                    'data_bits': 8,
                    'baud_rate': 57600,
                    'parity': constants.Parity.none,
                    'stop_bits': constants.StopBits.one,
                    #'flow_control': constants.VI_ASRL_FLOW_NONE,
                    'flow_control': constants.VI_ASRL_FLOW_XON_XOFF,  # constants.VI_ASRL_FLOW_NONE,
                    },
                }

    def __init__(self, *args, **kwargs):
        self.motionaxis_class = kwargs.pop('motionaxis_class', MotionAxisSMC100)
        super().__init__(*args, **kwargs)

    def initialize(self):
        super().initialize()

        # Clear read buffer
        self.clear_read_buffer()

        self.detect_axis()

    @Action()
    def clear_read_buffer(self):
        '''Read all data that was still in the read buffer and discard this'''
        try:
            while True:
                self.read()
        except pyvisa.errors.VisaIOError:
            pass  # readbuffer was empty already

    @Action()
    def detect_axis(self):
        """ Find the number of axis available.

        The detection stops as soon as an empty controller is found.
        """
        self.axes = []
        i = 0
        scan_axes = True
        while scan_axes:
            i += 1
            try:
                idn = self.query('%dID?' % i)
            except pyvisa.errors.VisaIOError:
                scan_axes = False
            else:
                if idn == '':
                    scan_axes = False
                else:
                    axis = self.motionaxis_class(self, i, idn)
                    self.axes.append(axis)

            

    
class MotionAxisSMC100(MotionAxis):
    def query(self, command, *, send_args=(None, None),
              recv_args=(None, None)):
        respons = super().query(command, send_args=send_args,
                                recv_args=recv_args)
        # check for command:
        if not respons[:3] == '{:d}{}'.format(self.num, command[:2]):
            self.log_error('Axis {}: Expected to return command {} instead of'
                           '{}'.format(self.num, command[:3], respons[:3]))
        return respons[3:]

    def write(self, command, *args, **kwargs):
        super().write(command, *args, **kwargs)
        return self.get_errors()

    @Feat(units='mm')
    def software_limit_positive(self):
        '''Make sure that software limits are tighter than hardware limits,
        else the stage will go to not reference mode'''
        return self.query('SR?')

    @software_limit_positive.setter
    def software_limit_positive(self, val):
        return self._software_limit_setter(val, limit='positive')

    @Feat(units='mm')
    def software_limit_negative(self):
        return self.query('SL?')

    @software_limit_negative.setter
    def software_limit_negative(self, val):
        return self._software_limit_setter(val, limit='negative')

    def _software_limit_setter(self, val, limit='positive'):
        self.enter_config_state()
        if limit == 'positive':
            ret = self.write('SR{}'.format(val))
        elif limit == 'negative':
            ret = self.write('SL{}'.format(val))
        else:
            self.log_error("Limit {} not in ('postive', 'negative')."
                           "".format(limit))
        self.leave_and_save_config_state()
        return ret

    @Action()
    def enter_config_state(self):
        return self.write('PW1')

    @Action()
    def leave_and_save_config_state(self):
        '''Takes up to 10s, controller is unresposive in that time'''
        super().write('PW0')
        start = time.time()
        # do-while loop
        cont = True
        while cont:
            try:
                self.status
            except ValueError:
                if (time.time() - start > 10):
                    self.log_error('Controller was going to CONFIGURATION '
                                   'state but it took more than 10s. Trying '
                                   'to continue anyway')
                    cont = False
                else:
                    time.sleep(0.001)
            else:
                cont = False

    @Action()
    def on(self):
        """Put axis on"""
        pass
        self.write('MM1')

    @Action()
    def off(self):
        """Put axis off"""
        pass
        self.write('MM0')

    @Action()
    def get_errors(self):
        ret = self.query('TE?')
        err = ERRORS.get(ret, 'Error {}. Lookup in manual: https://www.newpor'
                         't.com/medias/sys_master/images/images/h11/he1/91171'
                         '82525470/SMC100CC-SMC100PP-User-s-Manual.pdf'
                         ''.format(ret))
        if err:
            self.log_error('Axis {} error: {}'.format(self.num, err))
        return err

    @Feat()
    def status(self):
        '''Read and parse controller and axis status. This gives usefull error
        messages'''
        res = self.query('TS?')
        positioner_error = [val for key, val in positioner_errors.items() if
                            int(res[:4], base=16) & key == key]
        controller_state = controller_states[res[-2:]]
        return positioner_error, controller_state

    @Feat(values={True: '1', False: '0'})
    def is_on(self):
        """
        :return: True is axis on, else false
        """
        return '1'
        # return self.query('MM?')

    @Action()
    def home(self):
        super().home()
        self._wait_until_done()

    def _wait_until_done(self):
        er, st = self.status
        if st == 'MOVING.':
            time.sleep(self.wait_time)
            return self._wait_until_done()
        elif st[:5] == 'READY':
            return True
        else:
            self.log_error('Not reached position. Controller state: {} '
                           'Positioner errors: {}'
                           ''.format(st, ','.join(er)))
            return False

    @Feat()
    def motion_done(self):
        if self.status[1][:5] == 'READY':
            return True
        return False

    @Action()
    def keypad_disable(self):
        return self.write('JD')


if __name__ == '__main__':
    import argparse
    import lantz.log

    parser = argparse.ArgumentParser(description='Test SMC100 driver')
    parser.add_argument('-p', '--port', type=str, default='1',
                        help='Serial port to connect to')

    args = parser.parse_args()
    lantzlog = lantz.log.log_to_screen(level=lantz.log.INFO)
    lantz.log.log_to_socket(lantz.log.DEBUG)

    import lantz
    import visa
    import lantz.drivers.newport_motion
    sm = lantz.drivers.newport_motion.SMC100
    rm = visa.ResourceManager('@py')
    lantz.messagebased._resource_manager = rm

    print(lantz.messagebased._resource_manager.list_resources())

    with sm(args.port) as inst:
    #with sm.via_serial(port=args.port) as inst:
        inst.idn
        # inst.initialize() # Initialize the communication with the power meter
        # Find the status of all axes:
        #for axis in inst.axes:
        #    print('Axis {} Position {} is_on {} max_velocity {} velocity {}'.format(axis.num, axis.position,
        #                                                                            axis.is_on, axis.max_velocity,
        #                                                                            axis.velocity))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.signalrecovery
    ~~~~~~~~~~~~~~~~~~~~~

    :company: Signal Recovery
    :description: Measurement Technologies
    :website: http://www.signalrecovery.com/

    ---

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD,

"""

from .sr7265 import SR7265

__all__ = ['SR7265']



# -*- coding: utf-8 -*-
"""
    lantz.drivers.labjack
    ~~~~~~~~~~~~~~~~~~~~~

    :company: LabJack
    :description: LabJacks are USB/Ethernet based measurement and automation
                  devices which provide analog inputs/outputs, digital inputs/outputs, and more.

    :website: http://www.labjack.com/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .u12 import U12

__all__ = ['U12']




from lantz.messagebased import MessageBasedDriver
from lantz import Feat, Q_

class BNC645(MessageBasedDriver):

    DEFAULTS = {
        'COMMON': {
            'write_termination': '\n',
            'read_termination': '\n',
        },
    }

    # @Feat(units='V')
    # def amplitude(self):
    #     value = self.query("VOLT?")
    #     return float(value)
    #
    # @amplitude.setter
    # def amplitude(self, value):
    #     self.write("VOLT {}".format(value))
    #     return

    @Feat()
    def amplitude(self):
        self.write("VOLT:UNIT DBM")
        value = self.query("VOLT?")
        return float(value)

    @amplitude.setter
    def amplitude(self, value):
        self.write("VOLT:UNIT DBM")
        self.write("VOLT {}".format(value))
        return

    @Feat(units='Hz')
    def frequency(self):
        value = self.query("FREQ?")
        return float(value)

    @frequency.setter
    def frequency(self, value):
        self.write("FREQ {:1.8e}".format(value))
        return

    @Feat(values={True: '1', False: '0'})
    def output(self):
        value = self.query("OUTP?")
        return value

    @output.setter
    def output(self, value):
        self.write("OUTP {}".format(value))
        return



# -*- coding: utf-8 -*-
"""
    lantz.drivers.ieee4882
    ~~~~~~~~~~~~~~~~~~~~~~~

    Implements a IEEE-488.2 commands.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz import Action, Feat, Driver


class IEEE4882Driver(Driver):
    """Implements mandatory functions for a IEE488.2 device.

    You can use it as a mixin class.
    """

    @Feat(read_once=True)
    def idn(self):
        """Instrument identification.
        """
        return self.parse_query('*IDN?',
                                format='{manufacturer:s},{model:s},{serialno:s},{softno:s}')

    @Feat(read_once=True)
    def fitted_options(self):
        """Fitted options.
        """
        return self.query('*OPT?').split(',')

    @Action()
    def reset(self):
        """Set the instrument functions to the factory default power up state.
        """
        self.send('*RST')

    @Action()
    def self_test(self):
        """Performs a complete instrument self-test.

        If test fails, one or more error messages will provide additional information.
        When available, Use SYSTem:ERRor? to read error queue.
        """
        return self.query('*TST?') == '0'

    @Action()
    def wait(self):
        """Inhibit execution of an overlapped command until the execution of
        the preceding operation has been completed.
        """
        self.send('*WAI')

    @Action()
    def trigger(self):
        """Equivalent to Group Execute Trigger.
        """
        self.send('*TRG')

    @Feat()
    def status_byte(self):
        """Status byte, a number between 0-255.

        Decimal sum of the bits in the register.
        Bit #6: Master Summary Status Bit (MSS)
                This bit is set, if one of the bits in STB becomes true
                and the corresponding bit in the SRE is enabled.
        Bit #5: Event Summary Bit (ESB)
                This bit is set, if one of the bits in ESR becomes true
                and the corresponding bit in the ESE is enabled.
        Bit #4: Message Available Bit (MAV)
                This bit is set, if there is a message in the output buffer available.
        """
        return int(self.query('*STB?'))

    @Feat()
    def service_request_enabled(self):
        """Service request enable register.

        Decimal sum of the bits in the register.
        """
        return int(self.query('*SRE?'))

    @service_request_enabled.setter
    def service_request_enabled(self, value):
        self.query('*SRE {0:d}'.format(value))

    event_status_reg = Feat()

    @event_status_reg.setter
    def event_status_reg(self):
        """Queries the event register for the Standard Event Register group.
        Register is read-only; bits not cleared when read.
        """
        return int(self.query('*ESR?'))

    @Feat()
    def event_status_enabled(self):
        """Enables bits in the enable register for the Standard Event Register group.
        The selected bits are then reported to bit 5 of the Status Byte Register.

        Decimal sum of the bits in the register.
        Bit #7: Enable ESB when Power on or restart
        Bit #5: Enable ESB when a Command Error occur
        Bit #3: Enable ESB when a Device Dependent Error occur
        Bit #0: Enable ESB when Operation complete

        Others are not used.
        """
        return int(self.query('*ESE?'))

    @event_status_enabled.setter
    def event_status_enabled(self, value):
        self.query('*ESE {0:d}', value)

    @Action()
    def clear_status(self):
        """Clears the event registers in all register groups.
         Also clears the error queue.
        """
        self.send('*CLS')

    @Action()
    def wait_operation_complete_bit(self):
        """Returns 1 to the output buffer after all pending commands complete.

        Other commands cannot be executed until this command completes.
        """
        return self.query('*OPC?')

    @Action()
    def set_operation_complete_bit(self):
        """Sets "Operation Complete" (bit 0) in the Standard Event register at
        the completion of the current operation.

        The purpose of this command is to synchronize your application with the instrument.
        Other commands may be executed before Operation Complete bit is set.
        """
        return self.query('*OPC')

    @Feat(values={True: 0, False: 1})
    def poweron_status_clear_enabled(self):
        """Enables or disables clearing of two specific registers at power on:
        - Standard Event enable register
        - Status Byte condition register
        - Questionable Data Register
        - Standard Operation Register
        """
        return self.query('*PSC?')

    @poweron_status_clear_enabled.setter
    def poweron_status_clear_enabled(self, value):
        self.query('*PSC {}'.format(value))

    @Action()
    def recall_state(self, location):
        """Recalls (*RCL) instrument state in specified non-volatile location.

        :param location: non-volatile storage location.
        """
        self.send('*RCL {}'.format(location))

    @Action()
    def save_state(self, location):
        """Saves instrument state in specified non-volatile location.

        Previously stored state in location is overwritten (no error is generated).
        :param location: non-volatile storage location.
        """
        self.send('*SAV'.format(location))



# -*- coding: utf-8 -*-
"""
    lantz.drivers.mpb
    ~~~~~~~~~~~~~~~~~

    :company: MPB Communications Inc.
    :description: Laser products.
    :website: http://www.mpbc.ca/

    ----

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .vfl import VFL

__all__ = ['VFL']



"""
    lantz.drivers.lantz_server.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This allows to access networked devices.abs

    WARNING: This is still experimental.  It uses pickle to serialize the object which is not secure in an untrusted environment.

    TODO: Implement DictFeat, Add Input check before executing, Fix autoquery when using autocomplete in ipython (maybe some caching is sufficient...)

    Author: Alexandre Bourassa
    Date: 05/24/2018
"""

import socketserver
import pickle
import codecs
from lantz import Driver, Q_, Feat, DictFeat, Action
import socket
import time
import sys
from importlib import import_module

def encode_data(data, protocol=None):
    data = pickle.dumps(data, protocol=protocol)
    if len(data) > 9999999999:
        raise OverflowError
    out = bytearray("{:010}".format(len(data)),'ascii')
    out.extend(data)
    return out

def msg_complete(data):
    if len(data)<10:
        return False
    length = int(data[:10])
    return len(data) == (length + 10)

def decode_data(data):
    if not msg_complete(data):
        raise Exception("Incomplete/Invalid data")
    length = data[:10]
    d = data[10:]
    obj = pickle.loads(d)

    # Special case to handle Quantity
    if repr(type(obj)) == "<class 'pint.quantity.build_quantity_class.<locals>.Quantity'>":
        obj = Q_(str(obj))
    return obj

def receive_all(recv_fun, timeout):
    start = time.time()
    data = bytearray()
    while (time.time()-start)<timeout:
        buf = recv_fun(1024)
        data.extend(buf)
        if msg_complete(data):
            return decode_data(data)
    raise TimeoutError


VALID_TYPES = {'Feat':Feat, 'Action':Action, 'DictFeat':DictFeat}
VALID_QUERY = ['SET', 'GET']
def build_query(property_type, property_name, query_type='GET',val=None, args=[], kwargs={}, key=None):
    if not property_type in VALID_TYPES:
        raise Exception('Invalid property type')
    if not query_type in VALID_QUERY:
        raise Exception('Invalid query type')
    if property_type == 'Feat':
        return {'property_type':property_type, 'property_name':property_name, 'query_type':query_type, 'val':val}
    if property_type == 'DictFeat':
        return {'property_type':property_type, 'property_name':property_name, 'query_type':query_type, 'val':val, 'args':args, 'kwargs':kwargs, 'key':key}
    if property_type == 'Action':
        return {'property_type':property_type, 'property_name':property_name, 'args':args, 'kwargs':kwargs}

def exec_query(dev, query):
    #TODO Add some input checks
    if query['property_type'] == 'Feat':
        if query['query_type'] == 'SET':
            try:
                setattr(dev, query['property_name'], query['val'])
                return build_reply()
            except:
                e = sys.exc_info()[0]
                print('error: \n{}'.format(e))
                return build_reply(error=e)
        elif query['query_type'] == 'GET':
            try:
                val = getattr(dev, query['property_name'])
                return build_reply(msg=val)
            except:
                e = sys.exc_info()[0]
                print('error: \n{}'.format(e))
                return build_reply(error=e)
    if query['property_type'] == 'Action':
        try:
            msg = getattr(dev, query['property_name'])(*query['args'], **query['kwargs'])
            return build_reply(msg=msg)
        except:
            e = sys.exc_info()[0]
            print('error: \n{}'.format(e))
            return build_reply(error=e)
    else:
        print('Unsupported')
        return build_reply(error='Unsupported')
    
def build_reply(error=None, msg=None):
    return {'error':error, 'msg':msg}
        


class Lantz_Server(socketserver.TCPServer):

    def __init__(self, host, port, device):
        class Lantz_Handler(socketserver.StreamRequestHandler):
            def handle(self):
                data = receive_all(self.request.recv, 1)
                print('received: {}'.format(data))
                reply = exec_query(device, data)
                print('reply: {}'.format(reply))
                self.request.sendall(encode_data(reply))

        super().__init__((host, port), Lantz_Handler)

# class Lantz_Base_Client(Driver):
#     def __init__(self, host, port, timeout=1):
#         self.host = host
#         self.port = port
#         self.timeout = timeout
                

#     def query(self, data):
#         #Initialize and send query
#         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
#         sock.connect((self.host, self.port))
#         sock.sendall(encode_data(data))

#         #Read back ans from the server
#         ans = receive_all(sock.recv, self.timeout)

#         sock.close()
#         return ans


class Device_Client():
    def __new__(cls, device_driver_class, host, port, timeout=1, allow_initialize_finalize=True):
        if type(device_driver_class) is str:
            class_name = device_driver_class.split('.')[-1]
            mod = import_module(device_driver_class.replace('.'+class_name, ''))
            device_driver_class = getattr(mod, class_name)
        
        class Device_Client_Instance(device_driver_class):
            __name__ = '_Device_Client.' + device_driver_class.__name__
            __qualname__ = 'Device_Client.' + device_driver_class.__name__
            _allow_initialize_finalize = allow_initialize_finalize
            def initialize(self):
                if self._allow_initialize_finalize:
                    self._initialize()
            def finalize(self):
                if self._allow_initialize_finalize:
                    self._finalize()

            def __init__(self, host, port, timeout=1):
                self.host = host
                self.port = port
                self.timeout = timeout
                

            def query(self, data):
                #Initialize and send query
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((self.host, self.port))
                sock.sendall(encode_data(data))

                #Read back ans from the server
                ans = receive_all(sock.recv, self.timeout)

                sock.close()
                return ans

        for feat_name, feat in device_driver_class._lantz_features.items():
            if isinstance(feat, Feat):
                def get_fun(_feat_name):
                    def f_(_self):
                        reply = _self.query(build_query('Feat', _feat_name, query_type='GET'))
                        if not reply['error'] is None:
                            raise reply['error']
                        else:
                            return reply['msg']
                    return f_

                def set_fun(_feat_name):
                    def f_(_self, val):
                        reply = _self.query(build_query('Feat', _feat_name, query_type='SET', val=val))
                        if not reply['error'] is None:
                            raise reply['error']
                        else:
                            return reply['msg']
                    return f_  
                setattr(Device_Client_Instance, feat_name, property(get_fun(feat_name), set_fun(feat_name)))
            #TODO implement DictFeat
            else:
                continue
        for action_name, action in device_driver_class._lantz_actions.items():
            def execute(_action_name):
                def f_(_self, *args, **kwargs):
                    reply = _self.query(build_query('Action', _action_name, args=args, kwargs=kwargs))
                    if not reply['error'] is None:
                        raise reply['error']
                    else:
                        return reply['msg']
                return f_
            if action_name in ['initialize', 'finalize']:
                setattr(Device_Client_Instance, '_'+action_name, execute(action_name))
            else:
                setattr(Device_Client_Instance, action_name, execute(action_name))

        
                    
        obj = Device_Client_Instance.__new__(Device_Client_Instance)
        obj.__init__(host, port, timeout=timeout)
        return obj

if __name__ == "__main__":
    HOST, PORT = "localhost", 9999

    from lantz.drivers.stanford.sg396 import SG396

    sg = SG396('tcpip::192.168.1.108')

    server = Lantz_Server(HOST, PORT, sg)

    server.serve_forever()
